import datetime
from collections import defaultdict
from typing import Any, Dict, Generator, List, Optional, Union

import numpy as np
import pandas as pd
from cehrbert.data_generators.hf_data_generator import UNKNOWN_VALUE
from cehrbert.data_generators.hf_data_generator.hf_dataset_mapping import (
    ED_VISIT_TYPE_CODES,
    INPATIENT_VISIT_TYPE_CODES,
    INPATIENT_VISIT_TYPES,
    DatasetMapping,
    VisitObject,
    get_value,
    has_events_and_get_events,
    replace_escape_chars,
)
from cehrbert.med_extension.schema_extension import Event
from cehrbert.runners.hf_runner_argument_dataclass import DataTrainingArguments
from cehrbert_data.const.artificial_tokens import (
    DISCHARGE_UNKNOWN_TOKEN,
    GENDER_UNKNOWN_TOKEN,
    RACE_UNKNOWN_TOKEN,
)
from cehrbert_data.const.common import NA
from cehrbert_data.decorators.patient_event_decorator_base import get_att_function
from datasets.formatting.formatting import LazyBatch
from dateutil.relativedelta import relativedelta
from pandas import Series

from cehrgpt.gpt_utils import (
    construct_age_sequence,
    construct_time_sequence,
    encode_demographics,
    multiple_of_10,
)
from cehrgpt.models.tokenization_hf_cehrgpt import (
    NONE_BIN,
    UNKNOWN_BIN,
    CehrGptTokenizer,
)

CEHRGPT_COLUMNS = [
    "concept_ids",
    "concept_value_masks",
    "number_as_values",
    "concept_as_values",
    "is_numeric_types",
    "concept_values",
    "units",
    "epoch_times",
    "ages",
]


def convert_date_to_posix_time(index_date: Union[datetime.date, int, float]) -> float:
    if isinstance(index_date, datetime.date):
        return (
            datetime.datetime.combine(index_date, datetime.datetime.min.time())
            .replace(tzinfo=datetime.timezone.utc)
            .timestamp()
        )
    elif isinstance(index_date, datetime.datetime):
        return index_date.replace(tzinfo=datetime.timezone.utc).timestamp()
    return index_date


class DatasetMappingDecorator(DatasetMapping):

    def batch_transform(
        self, records: Union[LazyBatch, Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Drop index_date if it contains None.

        :param records:
        :return:
        """
        if isinstance(records, LazyBatch):
            table = records.pa_table

            if "index_date" in table.column_names:
                index_col = table.column("index_date")
                if index_col.null_count > 0:
                    table = table.drop(["index_date"])
            records = LazyBatch(pa_table=table, formatter=records.formatter)
        else:
            if "index_date" in records:
                if pd.isna(records["index_date"][0]):
                    del records["index_date"]
        return super().batch_transform(records=records)

    def transform(self, record: Dict[str, Any]) -> Union[Dict[str, Any], Series]:
        raise NotImplemented("Must be implemented")


class MedToCehrGPTDatasetMapping(DatasetMappingDecorator):
    def __init__(
        self,
        data_args: DataTrainingArguments,
        include_inpatient_hour_token: bool = True,
    ):
        self._time_token_function = get_att_function(data_args.att_function_type)
        self._include_auxiliary_token = data_args.include_auxiliary_token
        self._inpatient_time_token_function = get_att_function(
            data_args.inpatient_att_function_type
        )
        self._include_demographic_prompt = data_args.include_demographic_prompt
        self._include_inpatient_hour_token = include_inpatient_hour_token

    """
    This mapping function converts the MED (https://github.com/Medical-Event-Data-Standard/meds/tree/main) extension
    to the CehrGPT format. We make several assumptions
    - The first event contains the demographic information
    - From the second event onward
        - the time of the event is visit_start_datetime.
        - the first measurement contains the code indicating a standard OMOP Visit concept_id (e.g. 9201, 9202)
        - in case of inpatient visits, the last measurement is assumed to
            contain the standard OMOP concept id for discharge facilities (e.g 8536)
        - in case of inpatient visits, datetime_value of the last measurement stores visit_end_datetime
    """

    def remove_columns(self):
        return ["patient_id", "visits", "birth_datetime"]

    @staticmethod
    def _update_cehrgpt_record(
        cehrgpt_record: Dict[str, Any],
        code: str,
        time: datetime.datetime,
        age: int,
        concept_value_mask: int = 0,
        number_as_value: float = 0.0,
        concept_as_value: str = "0",
        is_numeric_type: int = 0,
        unit: str = NA,
    ) -> None:
        cehrgpt_record["concept_ids"].append(replace_escape_chars(code))
        cehrgpt_record["ages"].append(age)
        cehrgpt_record["concept_value_masks"].append(concept_value_mask)
        cehrgpt_record["number_as_values"].append(number_as_value)
        cehrgpt_record["concept_as_values"].append(concept_as_value)
        cehrgpt_record["units"].append(unit)
        cehrgpt_record["is_numeric_types"].append(is_numeric_type)
        cehrgpt_record["epoch_times"].append(
            time.replace(tzinfo=datetime.timezone.utc).timestamp()
        )

    def transform(self, record: Dict[str, Any]) -> Dict[str, Any]:
        cehrgpt_record = {
            "person_id": record["patient_id"],
            "concept_ids": [],
            "ages": [],
            "concept_value_masks": [],
            "number_as_values": [],
            "concept_as_values": [],
            "units": [],
            "is_numeric_types": [],
            "epoch_times": [],
        }
        # Extract the demographic information
        birth_datetime = record["birth_datetime"]
        if isinstance(birth_datetime, pd.Timestamp):
            birth_datetime = birth_datetime.to_pydatetime()
        gender = record["gender"]
        gender = GENDER_UNKNOWN_TOKEN if gender == UNKNOWN_VALUE else gender
        race = record["race"]
        race = RACE_UNKNOWN_TOKEN if race == UNKNOWN_VALUE else race
        visits = record["visits"]
        # This indicates this is columnar format
        if isinstance(visits, dict):
            visits = sorted(self.convert_visit_columnar_to_python(visits))
        else:
            visits = sorted(visits, key=lambda _: get_value(_, "visit_start_datetime"))

        # Add the demographic tokens
        first_visit = visits[0]
        first_visit_start_datetime: datetime.datetime = get_value(
            first_visit, "visit_start_datetime"
        )
        starting_age = relativedelta(first_visit_start_datetime, birth_datetime).years
        year_str = f"year:{str(first_visit_start_datetime.year)}"
        age_str = f"age:{starting_age}"
        self._update_cehrgpt_record(
            cehrgpt_record, year_str, first_visit_start_datetime, starting_age
        )
        self._update_cehrgpt_record(
            cehrgpt_record, age_str, first_visit_start_datetime, starting_age
        )
        self._update_cehrgpt_record(
            cehrgpt_record, gender, first_visit_start_datetime, starting_age
        )
        self._update_cehrgpt_record(
            cehrgpt_record, race, first_visit_start_datetime, starting_age
        )

        # Use a data cursor to keep track of time
        datetime_cursor: Optional[datetime.datetime] = None
        visit: VisitObject
        # Loop through all the visits
        for i, visit in enumerate(visits):
            events: Generator[Event, None, None] = get_value(visit, "events")
            has_events, events = has_events_and_get_events(events)
            if not has_events:
                continue

            visit_start_datetime: datetime.datetime = get_value(
                visit, "visit_start_datetime"
            )
            # If visit_end_datetime is populated for the inpatient visit, we update the datetime_cursor
            visit_end_datetime: Optional[datetime.datetime] = get_value(
                visit, "visit_end_datetime"
            )

            # We assume the first measurement to be the visit type of the current visit
            visit_type = get_value(visit, "visit_type")
            is_er_or_inpatient = (
                visit_type in INPATIENT_VISIT_TYPES
                or visit_type in INPATIENT_VISIT_TYPE_CODES
                or visit_type in ED_VISIT_TYPE_CODES
            )

            # Add artificial time tokens to the patient timeline if timedelta exists
            if datetime_cursor is not None:
                time_delta = max((visit_start_datetime - datetime_cursor).days, 0)
                # This generates an artificial time token depending on the choice of the time token functions
                self._update_cehrgpt_record(
                    cehrgpt_record,
                    code=self._time_token_function(time_delta),
                    time=visit_start_datetime,
                    age=relativedelta(datetime_cursor, birth_datetime).years,
                )

            datetime_cursor = visit_start_datetime
            # Add a [VS] token
            self._update_cehrgpt_record(
                cehrgpt_record,
                code="[VS]",
                time=datetime_cursor,
                age=relativedelta(datetime_cursor, birth_datetime).years,
            )
            # Add a visit type token
            self._update_cehrgpt_record(
                cehrgpt_record,
                code=visit_type,
                time=datetime_cursor,
                age=relativedelta(datetime_cursor, birth_datetime).years,
            )
            # We need to insert an inpatient hour token right after the visit type, we calculate the hour interval
            # with respect to the midnight of the day
            if is_er_or_inpatient and self._include_inpatient_hour_token:
                if datetime_cursor.hour > 0:
                    # This generates an artificial time token depending on the choice of the time token functions
                    self._update_cehrgpt_record(
                        cehrgpt_record,
                        code=f"i-H{datetime_cursor.hour}",
                        time=datetime_cursor,
                        age=relativedelta(datetime_cursor, birth_datetime).years,
                    )

            # Keep track of the existing outpatient events, we don't want to add them again
            existing_duplicate_events = list()
            for e in events:
                # If the event doesn't have a time stamp, we skip it
                event_time: datetime.datetime = e["time"]
                if not event_time:
                    continue

                # If numeric_value exists, this is a concept/value tuple, we indicate this using a concept_value_mask
                numeric_value = e.get("numeric_value", None)
                text_value = e.get("text_value", None)
                # The unit might be populated with a None value
                unit = e.get("unit", NA) if e.get("unit", NA) else NA
                concept_value_mask = int(
                    numeric_value is not None or text_value is not None
                )
                if numeric_value is None and text_value is not None:
                    if text_value.isnumeric():
                        numeric_value = float(text_value)

                is_numeric_type = int(numeric_value is not None)
                code = replace_escape_chars(e["code"])

                # Create the event identity
                event_identity = (
                    (event_time, code, text_value, unit)
                    if is_er_or_inpatient
                    else (event_time.date(), code, text_value, unit)
                )

                # Add a medical token to the patient timeline
                # If this is an inpatient visit, we use the event time stamps to calculate age and date
                # because the patient can stay in the hospital for a period of time.
                if is_er_or_inpatient:
                    # Calculate the time diff in days w.r.t the previous measurement
                    time_diff_days = (event_time - datetime_cursor).days
                    # Update the datetime_cursor if the time diff between two neighboring measurements is greater than and
                    # equal to 1 day
                    if self._inpatient_time_token_function and time_diff_days > 0:
                        # This generates an artificial time token depending on the choice of the time token functions
                        self._update_cehrgpt_record(
                            cehrgpt_record,
                            code=f"i-{self._inpatient_time_token_function(time_diff_days)}",
                            time=event_time,
                            age=relativedelta(event_time, birth_datetime).years,
                        )

                    if self._include_inpatient_hour_token:
                        # if the time difference in days is greater than 0, we calculate the hour interval
                        # with respect to the midnight of the day
                        time_diff_hours = (
                            event_time.hour
                            if time_diff_days > 0
                            else int(
                                (event_time - datetime_cursor).total_seconds() // 3600
                            )
                        )

                        if time_diff_hours > 0:
                            # This generates an artificial time token depending on the choice of the time token functions
                            self._update_cehrgpt_record(
                                cehrgpt_record,
                                code=f"i-H{time_diff_hours}",
                                time=event_time,
                                age=relativedelta(event_time, birth_datetime).years,
                            )

                if event_identity in existing_duplicate_events:
                    continue

                self._update_cehrgpt_record(
                    cehrgpt_record,
                    code=code,
                    time=event_time,
                    age=relativedelta(event_time, birth_datetime).years,
                    concept_value_mask=concept_value_mask,
                    unit=unit,
                    number_as_value=numeric_value if numeric_value else 0.0,
                    concept_as_value=(
                        replace_escape_chars(text_value) if text_value else "0"
                    ),
                    is_numeric_type=is_numeric_type,
                )
                existing_duplicate_events.append(event_identity)
                # we only want to update the time stamp when data_cursor is less than the event time
                if datetime_cursor < event_time or datetime_cursor is None:
                    datetime_cursor = event_time
                    # We need to bound the datetime_cursor if the current visit is an admission type of visit
                    # as the associated events could be generated after the visits are complete
                    if is_er_or_inpatient and visit_end_datetime is not None:
                        datetime_cursor = min(datetime_cursor, visit_end_datetime)

            # For inpatient or ER visits, we want to discharge_facility to the end of the visit
            if is_er_or_inpatient:
                # If visit_end_datetime is populated for the inpatient visit, we update the datetime_cursor
                if visit_end_datetime is not None:
                    datetime_cursor = visit_end_datetime

                if self._include_auxiliary_token:
                    # Reuse the age and date calculated for the last event in the patient timeline for the discharge
                    # facility event
                    discharge_facility = get_value(visit, "discharge_facility")
                    if not discharge_facility:
                        discharge_facility = DISCHARGE_UNKNOWN_TOKEN
                    else:
                        discharge_facility = (
                            DISCHARGE_UNKNOWN_TOKEN
                            if discharge_facility == UNKNOWN_VALUE
                            else discharge_facility
                        )
                    self._update_cehrgpt_record(
                        cehrgpt_record,
                        code=discharge_facility,
                        time=datetime_cursor,
                        age=relativedelta(datetime_cursor, birth_datetime).years,
                    )

            # Reuse the age and date calculated for the last event in the patient timeline
            self._update_cehrgpt_record(
                cehrgpt_record,
                code="[VE]",
                time=datetime_cursor,
                age=relativedelta(datetime_cursor, birth_datetime).years,
            )

        # Generate the orders of the concepts that the cehrbert dataset mapping function expects
        cehrgpt_record["orders"] = list(
            range(1, len(cehrgpt_record["concept_ids"]) + 1)
        )

        # Add some count information for this sequence
        cehrgpt_record["num_of_concepts"] = len(cehrgpt_record["concept_ids"])
        cehrgpt_record["num_of_visits"] = len(visits)

        if record.get("index_date", None) is not None:
            cehrgpt_record["index_date"] = (
                record["index_date"].replace(tzinfo=datetime.timezone.utc).timestamp()
            )
        if record.get("label", None) is not None:
            cehrgpt_record["label"] = record["label"]
        if record.get("age_at_index", None) is not None:
            cehrgpt_record["age_at_index"] = record["age_at_index"]

        assert len(cehrgpt_record["epoch_times"]) == len(
            cehrgpt_record["concept_ids"]
        ), "The number of time stamps must match with the number of concepts in the sequence"

        return cehrgpt_record


class HFCehrGptTokenizationMapping(DatasetMappingDecorator):
    def __init__(
        self,
        concept_tokenizer: CehrGptTokenizer,
    ):
        self._concept_tokenizer = concept_tokenizer
        self._lab_token_ids = self._concept_tokenizer.lab_token_ids

    def remove_columns(self):
        return [
            "concept_value_masks",
            "is_numeric_types",
        ]

    def filter_out_invalid_tokens(self, record: Dict[str, Any]) -> Dict[str, Any]:
        column_names = []
        seq_length = len(record["concept_ids"])

        # We can't have "0" as a token in the tokenizer because it would break tokenization for "Race/0", "Visit/0"
        # This is a pre-caution
        if "0" in record["concept_ids"]:
            if isinstance(record["concept_ids"], np.ndarray):
                record["concept_ids"][record["concept_ids"] == "0"] = "Unknown"
            else:
                record["concept_ids"] = [
                    "Unknown" if x == "0" else x for x in record["concept_ids"]
                ]

        for k, v in record.items():
            if k not in CEHRGPT_COLUMNS:
                continue
            if isinstance(v, (list, np.ndarray)) and len(v) == seq_length:
                column_names.append(k)
        valid_concept_ids = self._concept_tokenizer.get_vocab().keys()
        valid_indices = [
            idx
            for idx, concept_id in enumerate(record["concept_ids"])
            if concept_id in valid_concept_ids
        ]
        if len(valid_indices) != len(record["concept_ids"]):
            for column in column_names:
                values = record[column]
                record[column] = [values[idx] for idx in valid_indices]
        return record

    def transform(self, record: Dict[str, Any]) -> Dict[str, Any]:
        # Reconstruct the ages input before the filter is applied
        record["ages"] = construct_age_sequence(
            record["concept_ids"], record.get("ages", None)
        )
        record["epoch_times"] = construct_time_sequence(
            record["concept_ids"], record.get("epoch_times", None)
        )
        # Remove the tokens from patient sequences that do not exist in the tokenizer
        record = self.filter_out_invalid_tokens(record)
        # If any concept has a value associated with it, we normalize the value
        record["input_ids"] = self._concept_tokenizer.encode(record["concept_ids"])
        assert len(record["input_ids"]) == len(record["concept_ids"]), (
            "The number of tokens must equal to the number of concepts\n"
            f"decoded concept_ids: {self._concept_tokenizer.decode(record['input_ids'], skip_special_tokens=False)}"
        )
        record["value_indicators"] = record["concept_value_masks"]
        if "number_as_values" not in record or "concept_as_values" not in record:
            record["number_as_values"] = [
                float(value) if isinstance(value, float) else None
                for value in record["concept_values"]
            ]
            record["is_numeric_types"] = [
                int(isinstance(value, float)) for value in record["concept_values"]
            ]
            record["concept_as_values"] = [
                value if isinstance(value, str) else None
                for value in record["concept_values"]
            ]
        if np.any(np.asarray(record["concept_value_masks"]) > 0):
            values = []
            for i, (
                concept_id,
                unit,
                concept_value_mask,
                number_as_value,
                concept_as_value,
                is_numeric_type,
            ) in enumerate(
                zip(
                    record["concept_ids"],
                    record["units"],
                    record["concept_value_masks"],
                    record["number_as_values"],
                    record["concept_as_values"],
                    record["is_numeric_types"],
                )
            ):
                if concept_value_mask == 1:
                    value = UNKNOWN_BIN
                    if is_numeric_type == 1:
                        if concept_id in self._concept_tokenizer.numeric_concept_ids:
                            value = self._concept_tokenizer.normalize(
                                concept_id, unit, number_as_value
                            )
                    elif isinstance(concept_as_value, str):
                        value = concept_as_value
                    values.append(value)
                else:
                    values.append(NONE_BIN)
            assert len(values) == len(record["input_ids"])
            record["values"] = self._concept_tokenizer.encode_value(values)
        else:
            record["values"] = self._concept_tokenizer.encode_value(
                [NONE_BIN for _ in range(len(record["concept_value_masks"]))]
            )
        # Delete these features because they contain null values and pyarrow cannot concatenate multiple records
        del record["number_as_values"]
        del record["concept_as_values"]
        return record


class HFFineTuningMapping(HFCehrGptTokenizationMapping):
    """Consider removing this transformation in the future."""

    def transform(self, record: Dict[str, Any]) -> Dict[str, Any]:
        record = super().transform(record)
        record.update(
            {
                "age_at_index": (
                    record["age"] if "age" in record else record["age_at_index"]
                ),
                "classifier_label": int(record["label"] > 0),
                "index_date": (
                    convert_date_to_posix_time(record["index_date"])
                    if "index_date" in record
                    else None
                ),
            }
        )
        return record

    def remove_columns(self):
        columns = super().remove_columns()
        columns.append("label")
        return columns


class ExtractTokenizedSequenceDataMapping:
    def __init__(
        self,
        person_index_date_map: Dict[int, List[Dict[str, Any]]],
        observation_window: int = 0,
    ):
        self.person_index_date_map = person_index_date_map
        self.observation_window = observation_window

    def _calculate_prediction_start_time(self, prediction_time: float):
        if self.observation_window and self.observation_window > 0:
            return max(prediction_time - self.observation_window * 24 * 3600, 0)
        return 0

    def transform(self, record: Dict[str, Any]) -> Dict[str, Any]:
        person_id = record["person_id"]
        prediction_times = self.person_index_date_map[person_id]
        prediction_start_end_times = [
            (
                self._calculate_prediction_start_time(
                    prediction_time_label_map["index_date"]
                    .replace(tzinfo=datetime.timezone.utc)
                    .timestamp()
                ),
                prediction_time_label_map["index_date"]
                .replace(tzinfo=datetime.timezone.utc)
                .timestamp(),
                prediction_time_label_map["label"],
            )
            for prediction_time_label_map in prediction_times
        ]
        observation_window_indices = np.zeros(
            (len(prediction_times), len(record["epoch_times"])), dtype=bool
        )
        for i, epoch_time in enumerate(record["epoch_times"]):
            for sample_n, (
                feature_extraction_time_start,
                feature_extraction_end_end,
                _,
            ) in enumerate(prediction_start_end_times):
                if (
                    feature_extraction_time_start
                    <= epoch_time
                    <= feature_extraction_end_end
                ):
                    observation_window_indices[sample_n][i] = True

        seq_length = len(record["epoch_times"])
        time_series_columns = ["concept_ids", "input_ids"]
        static_inputs = dict()
        for k, v in record.items():
            if k in ["concept_ids", "input_ids"]:
                continue
            if isinstance(v, (list, np.ndarray)) and len(v) == seq_length:
                time_series_columns.append(k)
            else:
                static_inputs[k] = v

        batched_samples = defaultdict(list)
        for (_, index_date, label), observation_window_index in zip(
            prediction_start_end_times, observation_window_indices
        ):
            for k, v in static_inputs.items():
                batched_samples[k].append(v)
            batched_samples["classifier_label"].append(label)
            batched_samples["index_date"].append(index_date)
            try:
                start_age = int(record["concept_ids"][1].split(":")[1])
            except Exception:
                start_age = -1
            batched_samples["age_at_index"].append(start_age)
            for time_series_column in time_series_columns:
                batched_samples[time_series_column].append(
                    np.asarray(record[time_series_column])[observation_window_index]
                )
        return batched_samples

    def batch_transform(self, record: Dict[str, Any]) -> Dict[str, Any]:
        all_batched_record = defaultdict(list)
        all_columns = record.keys()
        for i in range(len(record["concept_ids"])):
            one_record = {}
            for column in all_columns:
                one_record[column] = record[column][i]
            new_batched_record = self.transform(one_record)
            for k, v in new_batched_record.items():
                all_batched_record[k].extend(v)
        return all_batched_record



import argparse
import datetime
import glob
import os
import uuid
from datetime import timedelta
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
import pyarrow.parquet as pq
from tqdm import tqdm

from cehrgpt.generation.omop_entity import (
    ConditionOccurrence,
    Death,
    DrugExposure,
    Measurement,
    OmopEntity,
    Person,
    ProcedureOccurrence,
    VisitOccurrence,
)
from cehrgpt.gpt_utils import (
    extract_time_interval_in_days,
    generate_artificial_time_tokens,
    is_inpatient_att_token,
    is_visit_end,
    is_visit_start,
)
from cehrgpt.models.tokenization_hf_cehrgpt import END_TOKEN

# TODO: move these to cehrbert_data
STOP_TOKENS = ["VE", "[VE]", END_TOKEN]

OOV = "[OOV]"
CURRENT_PATH = Path(__file__).parent
START_TOKEN_SIZE = 4
ATT_TIME_TOKENS = generate_artificial_time_tokens()
TABLE_LIST = [
    "person",
    "visit_occurrence",
    "condition_occurrence",
    "procedure_occurrence",
    "drug_exposure",
    "death",
    "measurement",
]
DISCHARGE_CONCEPT_LIST = [4216643, 4021968, 4146681, 4161979]
OOV_CONCEPT_MAP = {
    1525734: "Drug",
    779414: "Drug",
    722117: "Drug",
    722118: "Drug",
    722119: "Drug",
    905420: "Drug",
    1525543: "Drug",
}


def extract_gender_concept_id(gender_token: str) -> int:
    if gender_token.startswith("Gender/"):
        return int(gender_token[len("Gender/") :])
    elif gender_token.isnumeric():
        return int(gender_token)
    else:
        return 0


def extract_race_concept_id(race_token: str) -> int:
    if race_token.startswith("Race/"):
        return int(race_token[len("Race/") :])
    elif race_token.isnumeric():
        return int(race_token)
    else:
        return 0


def create_folder_if_not_exists(output_folder, table_name):
    if not os.path.isdir(Path(output_folder) / table_name):
        os.mkdir(Path(output_folder) / table_name)


def generate_omop_concept_domain(concept_parquet) -> Dict[int, str]:
    """
    Generate a dictionary of concept_id to domain_id.

    :param concept_parquet: concept dataframe read from parquet file
    :return: dictionary of concept_id to domain_id
    """
    domain_dict = {}
    for i in concept_parquet.itertuples():
        domain_dict[i.concept_id] = i.domain_id
    return domain_dict


def generate_lab_stats_mapping(
    all_lab_stats: Optional[List[Dict[str, Any]]]
) -> Dict[int, Dict[str, Any]]:
    lab_stats_mapping = {}
    if all_lab_stats is not None:
        for lab_stats in all_lab_stats:
            # TODO: the numeric check will not hold true if we concatenate
            #  the concept with the corresponding unit concept
            if lab_stats["concept_id"].isnumeric():
                concept_id = int(lab_stats["concept_id"])
                count = lab_stats["concept_id"]
                if (concept_id in lab_stats_mapping) and (
                    count > lab_stats_mapping[concept_id]["count"]
                ):
                    lab_stats_mapping[concept_id] = {
                        "mean": lab_stats["mean"],
                        "std": lab_stats["std"],
                        "count": lab_stats["count"],
                    }
                else:
                    lab_stats_mapping[concept_id] = {
                        "mean": lab_stats["mean"],
                        "std": lab_stats["std"],
                        "count": lab_stats["count"],
                    }
    return lab_stats_mapping


def append_to_dict(
    export_dict: Dict[str, Dict[int, OmopEntity]],
    omop_entity: OmopEntity,
    entity_id: int,
):
    if omop_entity.get_table_name() not in export_dict:
        export_dict[omop_entity.get_table_name()] = {}
    export_dict[omop_entity.get_table_name()][entity_id] = omop_entity


def delete_bad_sequence(
    export_dict: Dict[str, Dict[int, OmopEntity]],
    id_mappings: Dict[str, Dict[int, int]],
    person_id: int,
):
    for table_name, id_mapping in id_mappings.items():
        omop_id_mapping = np.array(list(id_mapping.keys()))
        person_id_mapping = np.array(list(id_mapping.values()))
        ids_to_delete = omop_id_mapping[np.where(person_id_mapping == person_id)]
        for id in ids_to_delete:
            export_dict[table_name].pop(id)


def export_and_clear(
    output_folder: str,
    export_dict: Dict[str, Dict[int, OmopEntity]],
    export_error: Dict[str, Dict[str, str]],
    id_mappings_dict: Dict[str, Dict[int, int]],
    pt_seq_dict: Dict[int, str],
    is_parquet: bool = True,
):
    for table_name, records_to_export in export_dict.items():

        records_in_json = []
        # If there is no omop_entity, we skip it
        if len(export_dict[table_name]) == 0:
            continue

        for entity_id, omop_entity in export_dict[table_name].items():
            try:
                records_in_json.append(omop_entity.export_as_json())
            except AttributeError:
                # append patient sequence to export error list using pt_seq_dict.
                if table_name not in export_error:
                    export_error[table_name] = []
                person_id = id_mappings_dict[table_name][entity_id]
                export_error[table_name].append(pt_seq_dict[person_id])
                continue
        schema = next(iter(records_to_export.items()))[1].get_schema()
        output_folder_path = Path(output_folder)
        file_path = output_folder_path / table_name / f"{uuid.uuid4()}.parquet"
        table_df = pd.DataFrame(records_in_json, columns=schema)

        if is_parquet:
            table_df.to_parquet(file_path)
        else:
            table_df.to_csv(file_path, header=schema, index=False)

        export_dict[table_name].clear()


def _is_none(x):
    return x is None or np.isnan(x)


def get_num_records(parquet_files: List[str]):
    total = 0
    for file_path in parquet_files:
        parquet_file = pq.ParquetFile(file_path)
        total += parquet_file.metadata.num_rows
    return total


def record_generator(parquet_files):
    for file_path in parquet_files:
        df = pd.read_parquet(file_path)
        for record in df.itertuples():
            yield record


def gpt_to_omop_converter_batch(
    const: int,
    patient_sequence_parquet_files: List[str],
    domain_map: Dict[int, str],
    output_folder: str,
    buffer_size: int,
    use_original_person_id: bool,
):
    omop_export_dict = {}
    error_dict = {}
    export_error = {}
    id_mappings_dict = {}
    pt_seq_dict = {}

    for tb in TABLE_LIST:
        create_folder_if_not_exists(output_folder, tb)
        id_mappings_dict[tb] = {}

    visit_occurrence_id: int = const + 1
    condition_occurrence_id: int = const + 1
    procedure_occurrence_id: int = const + 1
    drug_exposure_id: int = const + 1
    measurement_id: int = const + 1

    # Default the person_id
    person_id: int = const + 1

    patient_record_generator = record_generator(patient_sequence_parquet_files)
    total_record = get_num_records(patient_sequence_parquet_files)

    for index, record in tqdm(enumerate(patient_record_generator), total=total_record):
        bad_sequence = False
        # If original_person_id is set to true, we retrieve it from the record.
        # If person_id doest not exist in the record, we use the default_person_id
        if use_original_person_id:
            person_id = getattr(record, "person_id", person_id)

        # Retrieve the
        concept_ids = getattr(record, "concept_ids")
        is_numeric_types = getattr(record, "is_numeric_types", None)
        number_as_values = getattr(record, "number_as_values", None)
        concept_as_values = getattr(record, "concept_as_values", None)
        units = getattr(record, "units", None)

        # Skip the start token if it is the first token
        if "start" in concept_ids[0].lower():
            concept_ids = concept_ids[1:]
            if is_numeric_types is not None:
                is_numeric_types = is_numeric_types[1:]
            if number_as_values is not None:
                number_as_values = number_as_values[1:]
            if concept_as_values is not None:
                concept_as_values = concept_as_values[1:]
            if units is not None:
                units = units[1:]

        clinical_events = concept_ids[START_TOKEN_SIZE:]
        # Skip the sequences whose sequence length is 0
        if len(clinical_events) == 0:
            continue
        # Skip the patients whose last token is not a valid end token
        if clinical_events[-1] not in STOP_TOKENS:
            continue

        is_numeric_types = (
            is_numeric_types[START_TOKEN_SIZE:]
            if is_numeric_types is not None and not np.all(pd.isna(is_numeric_types))
            else None
        )
        number_as_values = (
            number_as_values[START_TOKEN_SIZE:]
            if number_as_values is not None and not np.all(pd.isna(number_as_values))
            else None
        )
        concept_as_values = (
            concept_as_values[START_TOKEN_SIZE:]
            if concept_as_values is not None and not np.all(pd.isna(concept_as_values))
            else None
        )
        units = (
            units[START_TOKEN_SIZE:]
            if units is not None and not np.all(pd.isna(units))
            else None
        )

        # TODO:Need to decode if the input is tokenized
        [start_year, start_age, start_gender, start_race] = concept_ids[
            0:START_TOKEN_SIZE
        ]
        if "year" not in start_year.lower():
            continue

        try:
            start_year = start_year.split(":")[1]
            start_age = start_age.split(":")[1]
            birth_year = int(start_year) - int(start_age)
        except Exception as e:
            print(
                f"Failed to convert {concept_ids[0:START_TOKEN_SIZE]} due to {e}, skipping to the next record"
            )
            continue

        # Skip the patients whose birth year is either before 1900 or after this year
        if int(birth_year) < 1900 or int(birth_year) > datetime.date.today().year:
            continue

        p = Person(
            person_id=person_id,
            gender_concept_id=extract_gender_concept_id(start_gender),
            year_of_birth=birth_year,
            race_concept_id=extract_race_concept_id(start_race),
        )

        append_to_dict(omop_export_dict, p, person_id)
        id_mappings_dict["person"][person_id] = person_id
        pt_seq_dict[person_id] = " ".join(concept_ids)
        discharged_to_concept_id = 0
        date_cursor = datetime.datetime(year=int(start_year), month=1, day=1)
        vo = None
        inpatient_visit_indicator = False

        for event_idx, event in enumerate(clinical_events, 0):
            if event == OOV:
                continue
            # For bad sequences, we don't proceed further and break from the for loop
            if bad_sequence:
                break
            if is_visit_start(event):
                if event_idx == len(clinical_events) - 1:
                    break
                elif clinical_events[event_idx + 1] == "[DEATH]":
                    # If the [DEATH] token is not placed at the end of the sequence, this is a bad sequence
                    if event_idx + 2 != len(clinical_events) - 1:
                        bad_sequence = True
                        break
                    death = Death(p, date_cursor.date())
                    append_to_dict(omop_export_dict, death, person_id)
                    id_mappings_dict["death"][person_id] = person_id
                else:
                    try:
                        if clinical_events[event_idx + 1].startswith("Visit/"):
                            visit_concept_id = int(
                                clinical_events[event_idx + 1][len("Visit/") :]
                            )
                        else:
                            visit_concept_id = int(clinical_events[event_idx + 1])
                        inpatient_visit_indicator = visit_concept_id in [
                            9201,
                            262,
                            8971,
                            8920,
                        ]
                        if visit_concept_id in domain_map:
                            if (
                                domain_map[visit_concept_id] != "Visit"
                                and visit_concept_id != 0
                            ):
                                bad_sequence = True
                                break
                        else:
                            bad_sequence = True
                            break

                    except (IndexError, ValueError):
                        error_dict[person_id] = {}
                        error_dict[person_id]["concept_ids"] = " ".join(concept_ids)
                        error_dict[person_id]["error"] = "Wrong visit concept id"
                        bad_sequence = True
                        continue

                    vo = VisitOccurrence(
                        visit_occurrence_id, visit_concept_id, date_cursor, p
                    )
                    append_to_dict(omop_export_dict, vo, visit_occurrence_id)
                    id_mappings_dict["visit_occurrence"][
                        visit_occurrence_id
                    ] = person_id
                    visit_occurrence_id += 1

            elif event in ATT_TIME_TOKENS:
                if event[0] == "D":
                    att_date_delta = int(event[1:])
                elif event[0] == "W":
                    att_date_delta = int(event[1:]) * 7
                elif event[0] == "M":
                    att_date_delta = int(event[1:]) * 30
                elif event == "LT":
                    att_date_delta = 365 * 3
                else:
                    att_date_delta = 0
                # Between visits, the date delta is simply calculated as the date difference
                date_cursor = date_cursor.replace(
                    hour=0, minute=0, second=0, microsecond=0
                )
                date_cursor = date_cursor + timedelta(days=att_date_delta)
            elif inpatient_visit_indicator and is_inpatient_att_token(event):
                inpatient_time_span_in_days = extract_time_interval_in_days(event)
                # Reset the data cursor to the start of the day before adding the num of days parsed out from the token
                date_cursor = date_cursor.replace(hour=0, minute=0, second=0)
                date_cursor = date_cursor + timedelta(days=inpatient_time_span_in_days)
            elif inpatient_visit_indicator and event.startswith("i-H"):
                # Handle hour tokens differently than the day tokens
                # The way we construct the inpatient hour tokens is that the sum of the consecutive
                # hour tokens cannot exceed the current day, so the data_cursor is bounded by a
                # theoretical upper limit
                upper_bound = date_cursor.replace(
                    hour=0, minute=0, second=0
                ) + timedelta(hours=23, minutes=59, seconds=59)
                hour_delta = int(event[3:])
                date_cursor = date_cursor + timedelta(hours=hour_delta)
                if date_cursor > upper_bound:
                    date_cursor = upper_bound
            elif is_visit_end(event):
                if vo is None:
                    bad_sequence = True
                    break
                # If it's a VE token, nothing needs to be updated because it just means the visit ended
                if inpatient_visit_indicator:
                    vo.set_discharged_to_concept_id(discharged_to_concept_id)
                    vo.set_visit_end_date(date_cursor)
                    # if the discharged_to_concept_id patient had died, the death record is created
                    if discharged_to_concept_id == 4216643:
                        death = Death(
                            p, date_cursor.date(), death_type_concept_id=32823
                        )
                        append_to_dict(omop_export_dict, death, person_id)
                        id_mappings_dict["death"][person_id] = person_id
                        # If death record is generated, we need to stop the sequence conversion
                        break
                else:
                    pass
            elif event in [
                "START",
                start_year,
                start_age,
                start_gender,
                start_race,
                "[DEATH]",
            ]:
                # If it's a start token, skip it
                pass
            elif event.endswith("/0"):
                # This should capture the concept such as Visit/0, Discharge/0
                pass
            else:
                try:
                    concept_id = int(event)
                    if (
                        concept_id not in domain_map
                        and concept_id not in OOV_CONCEPT_MAP
                    ):
                        error_dict[person_id] = {}
                        error_dict[person_id]["concept_ids"] = " ".join(concept_ids)
                        error_dict[person_id][
                            "error"
                        ] = f"No concept id found: {concept_id}"
                        bad_sequence = True
                        continue
                    else:
                        # If the current concept_id is 'Patient Died', this means it can only occur in the
                        # discharged_to_concept_id field, which indicates the current visit has to be an inpatient
                        # visit, this concept_id can only appear at the second last position
                        if concept_id == 4216643:
                            # If the current visit is not inpatient, reject the sequence
                            if not inpatient_visit_indicator:
                                bad_sequence = True
                                continue
                            # # If the current token is not the second last one of the sequence, reject because
                            # # death can only appear at the end of the sequence
                            # if idx + 1 != len(tokens_generated) - 1:
                            #     bad_sequence = True
                            #     continue
                            # we also enforce the rule where the sequence has to end on a VE token
                            if event_idx + 1 < len(
                                clinical_events
                            ) and not is_visit_end(clinical_events[event_idx + 1]):
                                bad_sequence = True
                                continue

                        if concept_id in domain_map:
                            domain = domain_map[concept_id]
                        elif concept_id in OOV_CONCEPT_MAP:
                            domain = OOV_CONCEPT_MAP[concept_id]
                        else:
                            domain = None

                        if domain == "Visit" or concept_id in DISCHARGE_CONCEPT_LIST:
                            discharged_to_concept_id = concept_id
                        elif domain == "Condition":
                            co = ConditionOccurrence(
                                condition_occurrence_id, concept_id, vo, date_cursor
                            )
                            append_to_dict(
                                omop_export_dict, co, condition_occurrence_id
                            )
                            id_mappings_dict["condition_occurrence"][
                                condition_occurrence_id
                            ] = person_id
                            condition_occurrence_id += 1
                        elif domain == "Procedure":
                            po = ProcedureOccurrence(
                                procedure_occurrence_id, concept_id, vo, date_cursor
                            )
                            append_to_dict(
                                omop_export_dict, po, procedure_occurrence_id
                            )
                            id_mappings_dict["procedure_occurrence"][
                                procedure_occurrence_id
                            ] = person_id
                            procedure_occurrence_id += 1
                        elif domain == "Drug":
                            de = DrugExposure(
                                drug_exposure_id, concept_id, vo, date_cursor
                            )
                            append_to_dict(omop_export_dict, de, drug_exposure_id)
                            id_mappings_dict["drug_exposure"][
                                drug_exposure_id
                            ] = person_id
                            drug_exposure_id += 1
                        elif domain == "Measurement":
                            number_as_value = (
                                number_as_values[event_idx]
                                if number_as_values is not None
                                else None
                            )
                            concept_as_value = (
                                concept_as_values[event_idx]
                                if concept_as_values is not None
                                else None
                            )
                            is_numeric_type = (
                                is_numeric_types[event_idx]
                                if is_numeric_types is not None
                                else None
                            )
                            unit = units[event_idx] if units is not None else None
                            m = Measurement(
                                measurement_id,
                                measurement_concept_id=concept_id,
                                is_numeric_type=is_numeric_type,
                                value_as_number=number_as_value,
                                value_as_concept_id=concept_as_value,
                                visit_occurrence=vo,
                                measurement_datetime=date_cursor,
                                unit_source_value=unit,
                            )
                            append_to_dict(omop_export_dict, m, measurement_id)
                            id_mappings_dict["measurement"][measurement_id] = person_id
                            measurement_id += 1

                except ValueError:
                    error_dict[person_id] = {}
                    error_dict[person_id]["concept_ids"] = " ".join(concept_ids)
                    error_dict[person_id]["error"] = f"Wrong concept id: {event}"
                    bad_sequence = True
                    continue
        if bad_sequence:
            delete_bad_sequence(omop_export_dict, id_mappings_dict, person_id)

        if not use_original_person_id:
            person_id += 1

        if index != 0 and index % buffer_size == 0:
            export_and_clear(
                output_folder,
                omop_export_dict,
                export_error,
                id_mappings_dict,
                pt_seq_dict,
            )

    # Final flush to the disk if there are still records in the cache
    export_and_clear(
        output_folder, omop_export_dict, export_error, id_mappings_dict, pt_seq_dict
    )

    with open(Path(output_folder) / "concept_errors.txt", "w") as f:
        error_dict["total"] = len(error_dict)
        f.write(str(error_dict))
    with open(Path(output_folder) / "export_errors.txt", "w") as f:
        total = 0
        for k, v in export_error.items():
            total += len(v)
        export_error["total"] = total
        f.write(str(export_error))


def main(args):
    all_parquet_files = glob.glob(
        os.path.join(args.patient_sequence_path, "*parquet"), recursive=True
    )
    if len(all_parquet_files) == 0:
        raise RuntimeError(f"No parquet files found in {args.patient_sequence_path}")

    print(
        f"There are total {len(all_parquet_files)} parquet files detected in {args.patient_sequence_path}."
    )
    if not os.path.exists(args.output_folder):
        Path(args.output_folder).mkdir(parents=True, exist_ok=True)

    batched_parquet_files = np.array_split(all_parquet_files, args.cpu_cores)
    concept_pd = pd.read_parquet(args.concept_path)
    domain_map = generate_omop_concept_domain(concept_pd)

    pool_tuples = []
    # TODO: Need to make this dynamic
    const = 10000000
    for i in range(1, args.cpu_cores + 1):
        pool_tuples.append(
            (
                const * i,
                batched_parquet_files[i - 1],
                domain_map,
                args.output_folder,
                args.buffer_size,
                args.use_original_person_id,
            )
        )

    with Pool(processes=args.cpu_cores) as p:
        p.starmap(gpt_to_omop_converter_batch, pool_tuples)
        p.close()
        p.join()

    return print("Done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Arguments for converting patient sequences to OMOP"
    )
    parser.add_argument(
        "--output_folder",
        dest="output_folder",
        action="store",
        help="The path for the output_folder",
        required=True,
    )
    parser.add_argument(
        "--concept_path",
        dest="concept_path",
        action="store",
        help="The path for your concept_path",
        required=True,
    )
    parser.add_argument(
        "--buffer_size",
        dest="buffer_size",
        action="store",
        type=int,
        help="The size of the batch",
        required=False,
        default=1024,
    )
    parser.add_argument(
        "--patient_sequence_path",
        dest="patient_sequence_path",
        action="store",
        help="The path for your patient sequence",
        required=True,
    )
    parser.add_argument(
        "--cpu_cores",
        dest="cpu_cores",
        type=int,
        action="store",
        help="The number of cpu cores to use for multiprocessing",
        required=False,
        default=1,
    )
    parser.add_argument(
        "--use_original_person_id",
        dest="use_original_person_id",
        action="store_true",
        help="Whether or not to use the original person id",
    )

    main(parser.parse_args())



import os
from datetime import datetime
from typing import Dict, List, Optional, Union

import numpy as np
import polars as pl
import torch
from cehrbert.data_generators.hf_data_generator.cache_util import CacheFileCollector
from cehrbert.data_generators.hf_data_generator.meds_utils import (
    create_dataset_from_meds_reader,
)
from cehrbert.runners.hf_runner_argument_dataclass import DataTrainingArguments
from cehrbert.runners.runner_util import (
    get_meds_extension_path,
    load_parquet_as_dataset,
)
from datasets import DatasetDict, concatenate_datasets, load_from_disk
from meds import held_out_split, subject_id_field, train_split, tuning_split
from transformers import TrainingArguments
from transformers.utils import logging

from cehrgpt.data.hf_cehrgpt_dataset_mapping import (
    ExtractTokenizedSequenceDataMapping,
    MedToCehrGPTDatasetMapping,
)
from cehrgpt.runners.hf_gpt_runner_argument_dataclass import CehrGPTArguments

LOG = logging.get_logger("transformers")


def get_torch_dtype(torch_dtype: Optional[str] = None) -> Union[torch.dtype, str]:
    if torch_dtype and hasattr(torch, torch_dtype):
        return getattr(torch, torch_dtype)
    return torch.float


def data_collate_fn(features, model_type: torch.dtype, collator):
    batch = collator(features)
    if model_type != torch.float32:
        for key, value in batch.items():
            # Only convert float32 tensors to bfloat16
            if isinstance(value, torch.Tensor) and value.dtype == torch.float32:
                batch[key] = value.to(model_type)
    return batch


def prepare_finetune_dataset(
    data_args: DataTrainingArguments,
    training_args: TrainingArguments,
    cehrgpt_args: CehrGPTArguments,
    cache_file_collector: Optional[CacheFileCollector] = None,
) -> DatasetDict:
    # If the data is in the MEDS format, we need to convert it to the CEHR-BERT format
    if data_args.is_data_in_meds:
        meds_extension_path = get_meds_extension_path(
            data_folder=data_args.cohort_folder,
            dataset_prepared_path=data_args.dataset_prepared_path,
        )
        try:
            LOG.info(
                f"Trying to load the MEDS extension from disk at {meds_extension_path}..."
            )
            dataset = load_from_disk(meds_extension_path)
            if data_args.streaming:
                if isinstance(dataset, DatasetDict):
                    dataset = {
                        k: v.to_iterable_dataset(
                            num_shards=training_args.dataloader_num_workers
                        )
                        for k, v in dataset.items()
                    }
                else:
                    dataset = dataset.to_iterable_dataset(
                        num_shards=training_args.dataloader_num_workers
                    )
        except Exception as e:
            LOG.warning(e)
            dataset = create_dataset_from_meds_reader(
                data_args=data_args,
                dataset_mappings=[
                    MedToCehrGPTDatasetMapping(
                        data_args=data_args,
                        include_inpatient_hour_token=cehrgpt_args.include_inpatient_hour_token,
                    )
                ],
                cache_file_collector=cache_file_collector,
            )
            if not data_args.streaming:
                dataset.save_to_disk(str(meds_extension_path))
                stats = dataset.cleanup_cache_files()
                LOG.info(
                    "Clean up the cached files for the cehrgpt dataset transformed from the MEDS: %s",
                    stats,
                )
                if cache_file_collector:
                    # Clean up the files created from the data generator
                    cache_file_collector.remove_cache_files()
                dataset = load_from_disk(str(meds_extension_path))

        train_set = dataset["train"]
        validation_set = dataset["validation"]
        test_set = dataset["test"]

        if cehrgpt_args.meds_repartition:
            train_val_set = concatenate_datasets([train_set, validation_set])
            if data_args.streaming and data_args.validation_split_num:
                train_val_set = train_val_set.shuffle(
                    buffer_size=10_000, seed=training_args.seed
                )
                train_set = train_val_set.skip(data_args.validation_split_num)
                validation_set = train_val_set.take(data_args.validation_split_num)
            elif data_args.validation_split_percentage:
                dataset = train_val_set.train_test_split(
                    test_size=data_args.validation_split_percentage,
                    seed=training_args.seed,
                )
                train_set = dataset["train"]
                validation_set = dataset["test"]
            else:
                raise RuntimeError(
                    f"Can not split the data. If streaming is enabled, validation_split_num needs to be "
                    f"defined, otherwise validation_split_percentage needs to be provided. "
                    f"The current values are:\n"
                    f"validation_split_percentage: {data_args.validation_split_percentage}\n"
                    f"validation_split_num: {data_args.validation_split_num}\n"
                    f"streaming: {data_args.streaming}"
                )
    else:
        train_set, validation_set, test_set = create_dataset_splits(
            data_args=data_args,
            cehrgpt_args=cehrgpt_args,
            seed=training_args.seed,
        )
    # Organize them into a single DatasetDict
    final_splits = DatasetDict(
        {"train": train_set, "validation": validation_set, "test": test_set}
    )
    return final_splits


def create_dataset_splits(
    data_args: DataTrainingArguments,
    cehrgpt_args: CehrGPTArguments,
    seed: int,
):
    """
    Creates training, validation, and testing dataset splits based on specified splitting strategies.

    This function splits a dataset into training, validation, and test sets, using either chronological,
    patient-based, or random splitting strategies, depending on the parameters provided in `data_args`.

    - **Chronological split**: Sorts by a specified date and splits based on historical and future data.
    - **Patient-based split**: Splits by unique patient IDs to ensure that patients in each split are distinct.
    - **Random split**: Performs a straightforward random split of the dataset.

    If `data_args.test_data_folder` is provided, a test set is loaded directly from it. Otherwise,
    the test set is created by further splitting the validation set based on `test_eval_ratio`.

    Parameters:
        data_args (DataTrainingArguments): A configuration object containing data-related arguments, including:
            - `data_folder` (str): Path to the main dataset.
            - `test_data_folder` (str, optional): Path to an optional test dataset.
            - `chronological_split` (bool): Whether to split chronologically.
            - `split_by_patient` (bool): Whether to split by unique patient IDs.
            - `validation_split_percentage` (float): Percentage of data to use for validation.
            - `test_eval_ratio` (float): Ratio of test to validation data when creating a test set from validation.
            - `preprocessing_num_workers` (int): Number of processes for parallel data filtering.
            - `preprocessing_batch_size` (int): Batch size for batched operations.
        cehrgpt_args (CehrGPTArguments): A configuration object containing CehrGPTArguments.
        seed (int): Random seed for reproducibility of splits.

    Returns:
        Tuple[Dataset, Dataset, Dataset]: A tuple containing:
            - `train_set` (Dataset): Training split of the dataset.
            - `validation_set` (Dataset): Validation split of the dataset.
            - `test_set` (Dataset): Test split of the dataset.

    Raises:
        FileNotFoundError: If `data_args.data_folder` or `data_args.test_data_folder` does not exist.
        ValueError: If incompatible arguments are passed for splitting strategies.

    Example Usage:
        data_args = DataTrainingArguments(
            data_folder="data/",
            validation_split_percentage=0.1,
            test_eval_ratio=0.2,
            chronological_split=True
        )
        train_set, validation_set, test_set = create_dataset_splits(data_args, seed=42)
    """
    dataset = load_parquet_as_dataset(data_args.data_folder)
    test_set = (
        None
        if not data_args.test_data_folder
        else load_parquet_as_dataset(data_args.test_data_folder)
    )

    # Patient-based split
    LOG.info("Using the split_by_patient strategy")
    unique_patient_ids = dataset.unique("person_id")
    np.random.seed(seed)
    np.random.shuffle(unique_patient_ids)
    train_end = int(
        len(unique_patient_ids) * (1 - data_args.validation_split_percentage)
    )
    LOG.info(f"There are {len(unique_patient_ids)} patients in total")

    train_patient_ids, val_patient_ids, test_patient_ids = None, None, None

    if cehrgpt_args.patient_splits_path:
        # If the patient_splits path is a folder (cehrgpt patient splits), then we assume that it contains a list of parquet files.
        if os.path.isdir(cehrgpt_args.patient_splits_path):
            patient_splits = pl.read_parquet(
                os.path.join(cehrgpt_args.patient_splits_path, "*.parquet")
            )
        else:
            # If the patient_splits path is a file, it must be a parquet file
            file_parts = os.path.splitext(cehrgpt_args.patient_splits_path)
            if not file_parts or not file_parts[-1].endswith("parquet"):
                raise ValueError(
                    f"{cehrgpt_args.patient_splits_path} is not a valid patient splits path."
                )
            patient_splits = pl.read_parquet(cehrgpt_args.patient_splits_path)

        if len(patient_splits) == 0:
            raise RuntimeError(
                f"The patient_splits at {cehrgpt_args.patient_splits_path} contains empty rows"
            )
        split_values = patient_splits.select("split").unique()["split"].to_list()
        is_split_meds_schema = np.all(
            [
                split in split_values
                for split in [train_split, tuning_split, held_out_split]
            ]
        )
        # Auto-detect whether or not the patient splits follow the MEDS schema or not.
        is_data_in_meds = (
            subject_id_field in patient_splits.columns
        ) and is_split_meds_schema
        patient_splits = patient_splits.filter(
            pl.col(subject_id_field if is_data_in_meds else "person_id").is_in(
                unique_patient_ids
            )
        )
        LOG.info(
            f"Patient Splits Schema: %s",
            "MEDS patient splits" if is_data_in_meds else "Traditional patient splits",
        )
        LOG.info("Effective patient_splits size: %s", len(patient_splits))
        if is_data_in_meds:
            train_patient_ids = patient_splits.filter(pl.col("split") == train_split)[
                subject_id_field
            ].to_list()
            val_patient_ids = patient_splits.filter(pl.col("split") == tuning_split)[
                subject_id_field
            ].to_list()
            test_patient_ids = patient_splits.filter(pl.col("split") == held_out_split)[
                subject_id_field
            ].to_list()
        else:
            train_patient_ids = patient_splits.filter(pl.col("split") == "train")[
                "person_id"
            ].to_list()
            val_patient_ids = patient_splits.filter(pl.col("split") == "validation")[
                "person_id"
            ].to_list()
            test_patient_ids = patient_splits.filter(pl.col("split") == "test")[
                "person_id"
            ].to_list()

    # In case that we could not retrieve the corresponding patient_ids
    if not train_patient_ids:
        train_patient_ids = unique_patient_ids[:train_end]

    if not val_patient_ids:
        val_patient_ids = unique_patient_ids[train_end:]

    if not test_set and not test_patient_ids:
        validation_end = int(len(val_patient_ids) * data_args.test_eval_ratio)
        val_patient_ids = val_patient_ids[:validation_end]
        test_patient_ids = val_patient_ids[validation_end:]

    # Helper function to apply patient-based filtering
    def filter_by_patient_ids(patient_ids):
        unique_ids = set(patient_ids)
        return dataset.filter(
            lambda batch: [pid in unique_ids for pid in batch["person_id"]],
            num_proc=data_args.preprocessing_num_workers,
            batched=True,
            batch_size=data_args.preprocessing_batch_size,
        )

    # Generate splits
    train_set = filter_by_patient_ids(train_patient_ids).shuffle(seed=seed)
    validation_set = filter_by_patient_ids(val_patient_ids)
    if not test_set:
        test_set = filter_by_patient_ids(test_patient_ids)

    LOG.info("Train size: %s", len(train_set))
    LOG.info("Validation size: %s", len(validation_set))
    LOG.info("Test size: %s", len(test_set))

    return train_set, validation_set, test_set


def extract_cohort_sequences(
    data_args: DataTrainingArguments,
    cehrgpt_args: CehrGPTArguments,
    cache_file_collector: Optional[CacheFileCollector] = None,
) -> DatasetDict:
    """
    Extracts and processes cohort-specific tokenized sequences from a pre-tokenized dataset,.

    based on the provided cohort Parquet files and observation window constraints.

    This function performs the following steps:
    1. Loads cohort definitions from Parquet files located in `data_args.cohort_folder`.
    2. Renames relevant columns if the data originates from a Meds format.
    3. Filters a pre-tokenized dataset (loaded from `cehrgpt_args.tokenized_full_dataset_path`)
       to include only patients present in the cohort.
    4. Aggregates each person's index date and label into a mapping.
    5. Checks for consistency to ensure all cohort person_ids are present in the tokenized dataset.
    6. Applies a transformation (`ExtractTokenizedSequenceDataMapping`) to generate
       observation-window-constrained patient sequences.
    7. Caches both the filtered and processed datasets using the provided `cache_file_collector`.

    Args:
        data_args (DataTrainingArguments): Configuration parameters for data processing,
            including cohort folder, observation window, batch size, and parallelism.
        cehrgpt_args (CehrGPTArguments): Contains paths to pre-tokenized datasets and CEHR-GPT-specific arguments.
        cache_file_collector (CacheFileCollector): Utility to register and manage dataset cache files.

    Returns:
        DatasetDict: A Hugging Face `DatasetDict` containing the processed datasets (e.g., train/validation/test),
                     where each entry includes sequences filtered and truncated by the observation window.

    Raises:
        RuntimeError: If any `person_id` in the cohort is missing from the tokenized dataset.
    """

    cohort = pl.read_parquet(os.path.join(data_args.cohort_folder, "*.parquet"))
    if data_args.is_data_in_meds:
        cohort = cohort.rename(
            mapping={
                "prediction_time": "index_date",
                "subject_id": "person_id",
                "boolean_value": "label",
            }
        )
    all_person_ids = cohort["person_id"].unique().to_list()
    # In case the label column does not exist, we add a fake column to the dataframe so subsequent process can work
    if "label" not in cohort.columns:
        cohort = cohort.with_columns(
            pl.Series(
                name="label", values=np.zeros_like(cohort["person_id"].to_numpy())
            )
        )

    # data_args.observation_window
    tokenized_dataset = load_from_disk(cehrgpt_args.tokenized_full_dataset_path)
    filtered_tokenized_dataset = tokenized_dataset.filter(
        lambda batch: [person_id in all_person_ids for person_id in batch["person_id"]],
        batched=True,
        batch_size=data_args.preprocessing_batch_size,
        num_proc=data_args.preprocessing_num_workers,
    )
    person_index_date_agg = cohort.group_by("person_id").agg(
        pl.struct("index_date", "label").alias("index_date_label")
    )
    # Convert to dictionary
    person_index_date_map: Dict[int, List[datetime]] = dict(
        zip(
            person_index_date_agg["person_id"].to_list(),
            person_index_date_agg["index_date_label"].to_list(),
        )
    )
    LOG.info(f"person_index_date_agg: {person_index_date_agg}")
    tokenized_person_ids = []
    for _, dataset in filtered_tokenized_dataset.items():
        tokenized_person_ids.extend(dataset["person_id"])
    missing_person_ids = [
        person_id
        for person_id in person_index_date_map.keys()
        if person_id not in tokenized_person_ids
    ]
    if missing_person_ids:
        raise RuntimeError(
            f"There are {len(missing_person_ids)} missing in the tokenized dataset. "
            f"The list contains: {missing_person_ids}"
        )
    processed_dataset = filtered_tokenized_dataset.map(
        ExtractTokenizedSequenceDataMapping(
            person_index_date_map, data_args.observation_window
        ).batch_transform,
        batched=True,
        batch_size=data_args.preprocessing_batch_size,
        num_proc=data_args.preprocessing_num_workers,
        remove_columns=filtered_tokenized_dataset["train"].column_names,
    )
    return processed_dataset



import unittest

from scipy.interpolate import UnivariateSpline

from cehrgpt.models.tokenization_hf_cehrgpt import create_bins_with_spline


class TestCreateBinsWithSpline(unittest.TestCase):
    def test_bins_count(self):
        samples = list(range(100))
        num_bins = 10
        bins = create_bins_with_spline(samples, num_bins)
        self.assertEqual(len(bins), num_bins)

    def test_bins_edges(self):
        samples = list(range(100))
        num_bins = 10
        bins = create_bins_with_spline(samples, num_bins)
        self.assertEqual(bins[0]["start_val"], float("-inf"))
        self.assertEqual(bins[-1]["end_val"], float("inf"))

    def test_few_samples(self):
        samples = list(
            range(5)
        )  # Not enough samples to form the requested number of bins
        num_bins = 10
        bins = create_bins_with_spline(samples, num_bins)
        self.assertEqual(len(bins), 0)

    def test_spline_fitting(self):
        samples = list(range(20))
        num_bins = 2
        bins = create_bins_with_spline(samples, num_bins)
        self.assertIsInstance(bins[0]["spline"], UnivariateSpline)
        self.assertIsInstance(bins[1]["spline"], UnivariateSpline)

    def test_spline_evaluation(self):
        samples = [x * x for x in range(10)]  # Quadratic relationship
        num_bins = 1
        bins = create_bins_with_spline(samples, num_bins)
        spline = bins[0]["spline"]
        x = 5
        # Check if the spline at mid-point roughly equals the mid-point y-value in samples
        self.assertAlmostEqual(spline(x), samples[x], delta=1.0)


if __name__ == "__main__":
    unittest.main()



import random
import unittest

from cehrgpt.gpt_utils import (
    RandomSampleCache,
    convert_time_interval_to_time_tuple,
    extract_time_interval_in_days,
    generate_artificial_time_tokens,
    is_att_token,
    is_inpatient_att_token,
    is_visit_start,
    random_slice_gpt_sequence,
)


class TestRandomSampleCache(unittest.TestCase):
    def test_cache_sampling_without_weights(self):
        cache = RandomSampleCache([1, 2, 3], cache_size=2)
        sample_1 = cache.next()
        sample_2 = cache.next()
        self.assertIn(sample_1, [1, 2, 3])
        self.assertIn(sample_2, [1, 2, 3])

    def test_cache_sampling_with_weights(self):
        cache = RandomSampleCache(
            [1, 2, 3], cache_size=2, sample_weights=[0.1, 0.1, 0.8]
        )
        sample = cache.next()
        self.assertIn(sample, [1, 2, 3])

    def test_cache_refill(self):
        cache = RandomSampleCache([1], cache_size=2)
        cache.next()
        self.assertEqual(len(cache._cache), 1)


class TestRandomSliceGPTSequence(unittest.TestCase):
    def test_random_slice_gpt_sequence_with_empty_starting_points(self):
        result = random_slice_gpt_sequence(["year:2020", "age:30", "male", "race"], 10)
        self.assertEqual((0, 0, ["year:2020", "age:30", "male", "race"]), result)

    def test_random_slice_gpt_sequence_with_valid_data(self):

        random.seed(0)

        # The pat sequence is greater than max_seq_len,
        # it should randomly pick a starting point and do not update the demographic prompt
        # because the time interval between the visits is 7 days
        concept_ids = [
            "year:2020",
            "age:30",
            "male",
            "race",
            "VS",
            "C",
            "C",
            "VE",
            "D7",
            "VS",
            "C",
            "VE",
        ]
        result = random_slice_gpt_sequence(concept_ids, 10)
        self.assertTrue(result[0] == 4)
        self.assertTrue(result[1] == 7)
        self.assertEqual(result[2], ["year:2020", "age:30", "male", "race"])

        concept_ids = [
            "year:2020",
            "age:30",
            "male",
            "race",
            "VS",
            "C",
            "C",
            "VE",
            "D500",
            "VS",
            "C",
            "VE",
            "VS",
            "C",
            "C",
            "VE",
            "D500",
            "VS",
            "C",
            "VE",
            "VS",
            "C",
            "C",
            "VE",
            "D500",
            "VS",
            "C",
            "VE",
        ]

        # The pat sequence is greater than max_seq_len,
        # it should randomly pick a starting point and update the demographic prompt
        result = random_slice_gpt_sequence(concept_ids, 20)
        self.assertTrue(result[0] == 9)
        self.assertTrue(result[1] == 23)
        self.assertEqual(result[2], ["year:2021", "age:31", "male", "race"])

        # The pat sequence is shorter than max_seq_len, it should return the original demographic prompt
        concept_ids = ["year:2020", "age:30", "male", "race", "VS", "C", "C", "VE"]
        result = random_slice_gpt_sequence(concept_ids, 10)
        self.assertTrue(result[0] == 0)
        self.assertTrue(result[1] == 0)
        self.assertEqual(result[2], ["year:2020", "age:30", "male", "race"])

    def test_random_slice_gpt_sequence_with_invalid_data(self):
        concept_ids = ["year:2020", "age:30", "male", "race", "invalid"]
        result = random_slice_gpt_sequence(concept_ids, 10)
        self.assertEqual(result, (0, 0, ["year:2020", "age:30", "male", "race"]))


class TestTokenFunctions(unittest.TestCase):
    def test_is_visit_start(self):
        self.assertTrue(is_visit_start("VS"))
        self.assertFalse(is_visit_start("D1"))

    def test_is_att_token(self):
        self.assertTrue(is_att_token("D7"))
        self.assertTrue(is_att_token("VS-D7"))
        self.assertFalse(is_att_token("random"))

    def test_is_inpatient_att_token(self):
        self.assertTrue(is_inpatient_att_token("VS-D7"))
        self.assertTrue(is_inpatient_att_token("VS-D7-VE"))
        self.assertTrue(is_inpatient_att_token("i-D7"))
        self.assertTrue(is_inpatient_att_token("i-D7-VE"))
        self.assertFalse(is_inpatient_att_token("V-D7"))
        self.assertFalse(is_inpatient_att_token("D7-VE"))


class TestTimeIntervalFunctions(unittest.TestCase):
    def test_extract_time_interval_in_days(self):
        self.assertEqual(extract_time_interval_in_days("D7"), 7)
        self.assertEqual(extract_time_interval_in_days("W1"), 7)
        self.assertEqual(extract_time_interval_in_days("M1"), 30)
        self.assertEqual(extract_time_interval_in_days("Y1"), 365)

    def test_extract_time_interval_invalid(self):
        with self.assertRaises(ValueError):
            extract_time_interval_in_days("invalid")

    def test_convert_time_interval_to_time_tuple(self):
        result = convert_time_interval_to_time_tuple(400, is_inpatient=True)
        self.assertEqual(result, ("year:1", "month:1", "i-day:5"))

    def test_convert_time_interval_negative(self):
        with self.assertRaises(AssertionError):
            convert_time_interval_to_time_tuple(-10, is_inpatient=False)


class TestGenerateArtificialTimeTokens(unittest.TestCase):
    def test_generate_artificial_time_tokens(self):
        tokens = generate_artificial_time_tokens()
        self.assertTrue("D1" in tokens)
        self.assertTrue("W1" in tokens)
        self.assertTrue("M1" in tokens)
        self.assertTrue("LT" in tokens)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd

# Define race mapping
race_mapping = {
    "38003613": "8557",
    "38003610": "8557",
    "38003579": "8515",
    "44814653": "0",
}

# Invalid age groups
invalid_age_groups = [
    "age:100-110",
    "age:110-120",
    "age:120-130",
    "age:130-140",
    "age:140-150",
    "age:150-160",
    "age:160-170",
    "age:170-180",
    "age:180-190",
    "age:190-200",
    "age:640-650",
    "age:680-690",
    "age:730-740",
    "age:740-750",
    "age:890-900",
    "age:900-910",
    "age:-10-0",
]


def age_group_func(age_str):
    """
    Categorize an age into a 10-year age group.

    Args:
        age_str (str): A string containing the age in the format "age:XX".

    Returns:
        str: A string representing the 10-year age group "age:XX-XX".
    """
    age = int(age_str.split(":")[1])
    group_number = age // 10
    return f"age:{group_number * 10}-{(group_number + 1) * 10}"


def map_race(race):
    return race_mapping.get(race, race)


def main(args):
    # Load data
    patient_sequence = pd.read_parquet(args.patient_sequence)
    # Extract and preprocess demographics
    demographics = patient_sequence.concept_ids.apply(
        lambda concept_ids: concept_ids[:4]
    )
    patient_sequence["demographics"] = demographics
    year = demographics.apply(lambda concepts: concepts[0])
    age = demographics.apply(lambda concepts: concepts[1]).apply(age_group_func)
    gender = demographics.apply(lambda concepts: concepts[2])
    race = demographics.apply(lambda concepts: concepts[3])
    death = patient_sequence.concept_ids.apply(
        lambda concept_ids: int(concept_ids[-2] == "[DEATH]")
    )

    patient_sequence["year"] = year
    patient_sequence["age"] = age
    patient_sequence["gender"] = gender
    patient_sequence["race"] = race
    patient_sequence["death"] = death

    demographics = patient_sequence[
        ["person_id", "death", "year", "age", "gender", "race", "split"]
    ]
    demographics["race"] = demographics.race.apply(map_race)

    demographics_clean = demographics[
        (demographics.gender != "0") & (~demographics.age.isin(invalid_age_groups))
    ]
    patient_sequence_clean = patient_sequence[
        patient_sequence.person_id.isin(demographics_clean.person_id)
    ]

    # Calculate probabilities
    probs = (
        demographics_clean.groupby(["age"])["person_id"].count()
        / len(demographics_clean)
    ).reset_index()
    probs.rename(columns={"person_id": "prob"}, inplace=True)

    # Adjust probabilities
    np.random.seed(42)
    x = np.asarray(list(reversed(range(1, 11))))
    adjusted_probs = probs.prob * pd.Series(x)
    adjusted_probs = adjusted_probs / adjusted_probs.sum()
    probs["adjusted_prob"] = adjusted_probs

    demographics_for_sampling = patient_sequence_clean[
        ["year", "age", "race", "gender", "person_id"]
    ].merge(probs, on="age")
    demographics_for_sampling["adjusted_prob"] = (
        demographics_for_sampling.adjusted_prob
        / demographics_for_sampling.adjusted_prob.sum()
    )

    # Train/Validation Split
    causal_train_split = demographics_for_sampling.sample(
        args.num_patients, replace=False, weights="adjusted_prob", random_state=1
    )
    causal_train_split["split"] = "train"
    causal_val_split = demographics_for_sampling[
        ~demographics_for_sampling.person_id.isin(causal_train_split.person_id)
    ]
    causal_val_split["split"] = "validation"

    causal_train_val_split = pd.concat([causal_train_split, causal_val_split])

    # Save outputs
    causal_train_val_split.to_parquet(args.output_folder, index=False)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Arguments for a causal patient split by age groups"
    )
    parser.add_argument(
        "--patient_sequence",
        required=True,
    )
    parser.add_argument(
        "--num_patients",
        default=1_000_000,
        type=int,
        required=False,
    )
    parser.add_argument(
        "--output_folder",
        required=True,
    )
    # Call the main function with parsed arguments
    main(parser.parse_args())



import os

from pyspark.sql import SparkSession
from pyspark.sql import functions as f
from pyspark.sql import types as t
from tqdm import tqdm

VOCAB_SIZE = 17870
L = 0.001
SENSITIVE_MATCH_THRESHOLD = int(VOCAB_SIZE * L)

# This number came from the tokenizer, essentially all numeric tokens that represent valid OMOP concept ids
COMMON_ATTRIBUTES = [
    "320128",
    "200219",
    "77670",
    "432867",
    "254761",
    "312437",
    "378253",
]
MAX_LEVELS = [1, 0, 0] + [1] * len(COMMON_ATTRIBUTES)
# The age groups are generated using the following bases e.g. 1 indicates using the age as is
AGE_GENERALIZATION_LEVELS = [10, 1]


def generate_lattice_bfs(top_gen_levels):  # BFS
    """Came from https://github.com/yy6linda/synthetic-ehr-benchmarking/blob/main/privacy_evaluation/Synthetic_risk_model_reid.py#L81."""
    visited = [top_gen_levels]
    queue = [top_gen_levels]
    lattice = []
    while queue:
        gen_levels = queue.pop(0)
        lattice.append(gen_levels)
        for i in range(len(gen_levels)):
            if gen_levels[i] != 0:
                gen_levels_new = gen_levels.copy()
                gen_levels_new[i] -= 1
                if not gen_levels_new in visited:
                    visited.append(gen_levels_new)
                    queue.append(gen_levels_new)
    return lattice


def update_dataset(reid_data, real_sample, synthetic_reid_data, config):
    reid_data_dup = reid_data.alias("reid_data_dup")
    real_sample_dup = real_sample.alias("real_sample_dup")
    synthetic_reid_data_dup = synthetic_reid_data.alias("synthetic_reid_data_dup")
    common_attributes_to_remove = []

    for i in range(len(config)):
        # age group
        if i == 0:
            age_level_index = config[i]
            age_group_base = AGE_GENERALIZATION_LEVELS[age_level_index]
            reid_data_dup = reid_data_dup.withColumn(
                "age", f.ceil(f.col("age").cast("int") / age_group_base)
            )
            synthetic_reid_data_dup = synthetic_reid_data_dup.withColumn(
                "age", f.ceil(f.col("age").cast("int") / age_group_base)
            )
            real_sample_dup = real_sample_dup.withColumn(
                "age", f.ceil(f.col("age").cast("int") / age_group_base)
            )
        elif i in [1, 2]:
            # gender and race are not generalized
            continue
        else:
            # this indicates that the common attribute should be removed
            if config[i] == 0:
                common_attributes_to_remove.append(COMMON_ATTRIBUTES[i - 3])

    def remove_common_attributes(concept_ids):
        comm_atts = sorted(set(concept_ids) - set(common_attributes_to_remove))
        if comm_atts:
            return "-".join(comm_atts)
        else:
            return "empty"

    extract_common_attributes_udf = f.udf(
        lambda concept_ids: remove_common_attributes(concept_ids), t.StringType()
    )

    reid_data_dup = reid_data_dup.withColumn(
        "common_attributes", extract_common_attributes_udf("common_attributes")
    )

    real_sample_dup = real_sample_dup.withColumn(
        "common_attributes", extract_common_attributes_udf("common_attributes")
    )

    synthetic_reid_data_dup = synthetic_reid_data_dup.withColumn(
        "common_attributes", extract_common_attributes_udf("common_attributes")
    )
    return reid_data_dup, real_sample_dup, synthetic_reid_data_dup


def calculate_reid_risk_score(
    real_sample_dup,
    reid_data_dup,
    synthetic_reid_data_dup,
    lower_n,
    cap_n,
    lambda_val=0.23,
    num_salts=20,
):
    real_sample_dup = real_sample_dup.withColumn(
        "salt", (f.rand() * num_salts).cast("int")
    )

    reid_data_stats = reid_data_dup.groupby(
        "age", "gender", "race", "common_attributes"
    ).count()

    real_to_population_matches = real_sample_dup.join(
        reid_data_stats,
        (real_sample_dup["age"] == reid_data_stats["age"])
        & (real_sample_dup["gender"] == reid_data_stats["gender"])
        & (real_sample_dup["race"] == reid_data_stats["race"])
        & (
            real_sample_dup["common_attributes"] == reid_data_stats["common_attributes"]
        ),
    ).select("person_id", "count")

    # Alias the DataFrame for self join
    real_sample_stats = real_sample_dup.groupby(
        "age", "gender", "race", "common_attributes"
    ).count()

    real_to_real_matches = real_sample_dup.join(
        real_sample_stats,
        (real_sample_dup["age"] == real_sample_stats["age"])
        & (real_sample_dup["gender"] == real_sample_stats["gender"])
        & (real_sample_dup["race"] == real_sample_stats["race"])
        & (
            real_sample_dup["common_attributes"]
            == real_sample_stats["common_attributes"]
        ),
    ).select("person_id", "count")

    reid_data_step_one = (
        reid_data_dup.join(real_to_population_matches, "person_id")
        .withColumnRenamed("count", "upper_F_s")
        .join(real_to_real_matches, "person_id")
        .withColumnRenamed("count", "lower_f_s")
    )

    synthetic_reid_data_dup = synthetic_reid_data_dup.withColumn(
        "salt", f.explode(f.array([f.lit(x) for x in range(num_salts)]))
    )

    real_sample_dup = real_sample_dup.where(
        f.size("sensitive_attributes") >= SENSITIVE_MATCH_THRESHOLD
    )
    synthetic_reid_data_dup = synthetic_reid_data_dup.where(
        f.size("sensitive_attributes") >= SENSITIVE_MATCH_THRESHOLD
    )

    real_to_synthetic = (
        real_sample_dup.join(
            synthetic_reid_data_dup,
            (real_sample_dup["age"] == synthetic_reid_data_dup["age"])
            & (real_sample_dup["gender"] == synthetic_reid_data_dup["gender"])
            & (real_sample_dup["race"] == synthetic_reid_data_dup["race"])
            & (
                real_sample_dup["common_attributes"]
                == synthetic_reid_data_dup["common_attributes"]
            )
            & (real_sample_dup["salt"] == synthetic_reid_data_dup["salt"]),
        )
        .select(
            real_sample_dup["person_id"],
            real_sample_dup["sensitive_attributes"].alias("real_sensitive_attributes"),
            synthetic_reid_data_dup["sensitive_attributes"].alias(
                "synthetic_sensitive_attributes"
            ),
        )
        .withColumn(
            "n_of_sensitive_matches",
            f.size(
                f.array_intersect(
                    "real_sensitive_attributes", "synthetic_sensitive_attributes"
                )
            ),
        )
        .drop("real_sensitive_attributes", "synthetic_sensitive_attributes")
    )

    real_to_synthetic = (
        real_to_synthetic.groupby("person_id")
        .agg(f.max("n_of_sensitive_matches").alias("n_of_sensitive_matches"))
        .withColumn(
            "new_info",
            (f.col("n_of_sensitive_matches") > f.lit(SENSITIVE_MATCH_THRESHOLD)).cast(
                "int"
            ),
        )
    )

    reid_data_step_two = reid_data_step_one.join(
        real_to_synthetic, "person_id", how="left_outer"
    ).withColumn("new_info", f.coalesce(real_to_synthetic["new_info"], f.lit(0)))

    return (
        reid_data_step_two.withColumn(
            "A_term",
            1 / f.col("lower_f_s") * f.lit((1 + lambda_val) / 2) * f.col("new_info"),
        )
        .withColumn(
            "B_term",
            1 / f.col("upper_f_s") * f.lit((1 + lambda_val) / 2) * f.col("new_info"),
        )
        .select(
            (f.sum("A_term") / f.lit(cap_n)).alias("A_term"),
            (f.sum("B_term") / f.lit(lower_n)).alias("B_term"),
        )
    )


def main(args):
    all_configs = generate_lattice_bfs(MAX_LEVELS)

    N = None
    n = None
    excluded_sensitive_attributes = None

    for config in tqdm(all_configs, total=len(all_configs)):

        spark = SparkSession.builder.appName(
            f'Generate REID {"".join(map(str, config))}'
        ).getOrCreate()

        if sum(config) == 0:
            continue

        experiment_output_name = os.path.join(
            args.output_folder, f'{"".join(map(str, config))}.parquet'
        )
        if os.path.exists(experiment_output_name):
            continue

        @f.udf(t.ArrayType(t.StringType()))
        def extract_omop_concepts_udf(concept_ids):
            return list(sorted(set([_ for _ in concept_ids if str.isnumeric(_)])))

        def extract_common_attributes(concept_ids):
            commn_atts = set([c for c in concept_ids if c in COMMON_ATTRIBUTES])
            return list(commn_atts)

        extract_common_attributes_udf = f.udf(
            lambda concept_ids: extract_common_attributes(concept_ids),
            t.ArrayType(t.StringType()),
        )

        def extract_sensitive_attributes(concept_ids):
            return list(
                sorted(set([c for c in concept_ids if c not in COMMON_ATTRIBUTES]))
            )

        extract_sensitive_attributes_udf = f.udf(
            lambda concept_ids: extract_sensitive_attributes(concept_ids),
            t.ArrayType(t.StringType()),
        )

        training_data = spark.read.parquet(args.training_data_folder)
        evaluation_data = spark.read.parquet(args.evaluation_data_folder)
        training_data = training_data.select(
            "concept_ids", "num_of_concepts", "person_id"
        ).withColumn("is_real", f.lit(True))
        evaluation_data = evaluation_data.select(
            "concept_ids", "num_of_concepts", "person_id"
        ).withColumn("is_real", f.lit(False))
        data = training_data.unionByName(evaluation_data)

        reid_data = (
            data.withColumn("age", f.col("concept_ids")[1])
            .withColumn("age", f.split("age", ":")[1])
            .withColumn("gender", f.col("concept_ids")[2])
            .withColumn("race", f.col("concept_ids")[3])
            .withColumn("concept_ids", extract_omop_concepts_udf("concept_ids"))
            .withColumn(
                "common_attributes", extract_common_attributes_udf("concept_ids")
            )
            .withColumn(
                "sensitive_attributes", extract_sensitive_attributes_udf("concept_ids")
            )
            .drop("concept_ids")
        )

        real_sample = reid_data.where("is_real")

        if not N:
            N = reid_data.count()
        if not n:
            n = real_sample.count()

        if not excluded_sensitive_attributes:
            excluded_sensitive_attributes_df = (
                real_sample.select(
                    f.explode("sensitive_attributes").alias("sensitive_attribute")
                )
                .groupby("sensitive_attribute")
                .count()
                .withColumn("sensitive_attribute_prevalence", f.col("count") / n)
                .withColumn(
                    "is_majority", f.col("sensitive_attribute_prevalence") >= 0.5
                )
                .where("is_majority")
            )
            excluded_sensitive_attributes = [
                row.sensitive_attribute
                for row in excluded_sensitive_attributes_df.select(
                    "sensitive_attribute"
                ).collect()
            ]

        def filter_sensitive_attributes(concept_ids):
            return [c for c in concept_ids if c not in excluded_sensitive_attributes]

        filter_sensitive_attributes_udf = f.udf(
            lambda concept_ids: filter_sensitive_attributes(concept_ids),
            t.ArrayType(t.StringType()),
        )

        real_sample = real_sample.withColumn(
            "sensitive_attributes",
            filter_sensitive_attributes_udf("sensitive_attributes"),
        )

        synthetic_data = spark.read.parquet(args.synthetic_data_folder)

        synthetic_reid_data = (
            synthetic_data.withColumn("age", f.col("concept_ids")[1])
            .withColumn("age", f.split("age", ":")[1])
            .withColumn("gender", f.col("concept_ids")[2])
            .withColumn("race", f.col("concept_ids")[3])
            .withColumn("concept_ids", extract_omop_concepts_udf("concept_ids"))
            .withColumn(
                "common_attributes", extract_common_attributes_udf("concept_ids")
            )
            .withColumn(
                "sensitive_attributes", extract_sensitive_attributes_udf("concept_ids")
            )
            .drop("concept_ids")
        )

        reid_data_dup, real_sample_dup, synthetic_reid_data_dup = update_dataset(
            reid_data, real_sample, synthetic_reid_data, config=config
        )
        experiment_risk_score = calculate_reid_risk_score(
            real_sample_dup,
            reid_data_dup,
            synthetic_reid_data_dup,
            lower_n=n,
            cap_n=N,
            lambda_val=0.23,
            num_salts=args.num_salts,
        )
        experiment_risk_score.toPandas().to_parquet(experiment_output_name)

        spark.stop()


def create_argparser():
    import argparse

    parser = argparse.ArgumentParser(
        description="Arguments for re-identification risk evaluation"
    )
    parser.add_argument(
        "--training_data_folder",
        dest="training_data_folder",
        action="store",
        required=True,
    )
    parser.add_argument(
        "--evaluation_data_folder",
        dest="evaluation_data_folder",
        action="store",
        required=True,
    )
    parser.add_argument(
        "--synthetic_data_folder",
        dest="synthetic_data_folder",
        action="store",
        required=True,
    )
    parser.add_argument(
        "--output_folder",
        dest="output_folder",
        action="store",
        help="The output folder for storing the results",
        required=True,
    )
    parser.add_argument(
        "--num_salts",
        dest="num_salts",
        action="store",
        type=int,
        default=40,
        required=False,
    )
    return parser


if __name__ == "__main__":
    main(create_argparser().parse_args())



import math
import unittest

import torch

from cehrgpt.models.gpt2 import RotaryPositionEmbedding


class TestRotaryPositionEmbedding(unittest.TestCase):

    def test_known_2d_rotation(self):
        emb = RotaryPositionEmbedding(dim=2)
        x = torch.tensor([[1.0, 0.0]])
        time = torch.tensor([[0.5]])
        output = emb(x, time)
        theta = 0.5 * emb.inv_freq[0, 0, 0]
        expected = torch.tensor([[math.cos(theta), math.sin(theta)]])
        self.assertTrue(
            torch.allclose(output, expected, atol=1e-5),
            f"Expected {expected}, got {output}",
        )

    def test_norm_preservation(self):
        emb = RotaryPositionEmbedding(dim=8)
        x = torch.randn(10, 8)
        time = torch.randint(0, 10, size=(10, 1)).float()
        x_rot = emb(x, time)
        original_norm = torch.norm(x, dim=-1)
        rotated_norm = torch.norm(x_rot, dim=-1)
        self.assertTrue(torch.allclose(original_norm, rotated_norm, atol=1e-5))

    def test_relative_position_consistency(self):
        emb = RotaryPositionEmbedding(dim=2)
        x = torch.tensor([[1.0, 0.0]])
        t1 = torch.tensor([[1.0]])
        t2 = torch.tensor([[2.0]])
        t3 = torch.tensor([[3.0]])
        d1 = emb(x, t2) - emb(x, t1)
        d2 = emb(x, t3) - emb(x, t2)
        self.assertFalse(
            torch.allclose(d1, d2, atol=1e-3)
        )  # Should not be exactly equal
        self.assertTrue(torch.isfinite(d1 - d2).all())  # But valid difference

    def test_dot_product_relative_shift_invariance(self):
        dim = 8
        emb = RotaryPositionEmbedding(dim=dim)
        x = torch.randn(1, dim)
        y = torch.randn(1, dim)

        t1 = torch.tensor([[3.0]])
        t2 = torch.tensor([[7.0]])
        delta = 5.0

        # RoPE embeddings at original positions
        x1 = emb(x, t1)
        y1 = emb(y, t2)

        # RoPE embeddings at shifted positions
        x2 = emb(x, t1 + delta)
        y2 = emb(y, t2 + delta)

        dot1 = torch.sum(x1 * y1, dim=-1)
        dot2 = torch.sum(x2 * y2, dim=-1)

        self.assertTrue(
            torch.allclose(dot1, dot2, atol=1e-5),
            f"Dot products differ: {dot1.item():.6f} vs {dot2.item():.6f}",
        )

    def test_batched_inputs(self):
        emb = RotaryPositionEmbedding(dim=6)
        x = torch.randn(2, 6)
        time = torch.tensor([[1.0], [3.0]])
        x_rot = emb(x, time)
        self.assertEqual(x_rot.shape, x.shape)
        self.assertTrue(torch.isfinite(x_rot).all())

    def test_dtype_preservation(self):
        emb = RotaryPositionEmbedding(dim=4)
        x = torch.randn(5, 4, dtype=torch.float32)
        time = torch.randint(0, 10, (5, 1)).float()
        out = emb(x, time)
        self.assertEqual(out.dtype, x.dtype)


if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from src.cehrgpt.tools.upload_omop_tables import upload_omop_tables


class TestOMOPUpload(unittest.TestCase):

    @patch("src.cehrgpt.tools.upload_omop_tables.f")  # Mock Spark SQL functions
    @patch("src.cehrgpt.tools.upload_omop_tables.SparkSession")  # Mock SparkSession
    def test_upload_omop_tables(self, mock_spark, mock_f):
        # Mock the Spark session and DataFrame
        mock_df = MagicMock()
        mock_spark.read.format.return_value.load.return_value = mock_df

        # Mock columns for casting
        mock_df.columns = ["concept_id", "measurement_date"]

        # Mock `f.col()` and `f.lit()` to return other mocks
        mock_col = MagicMock()
        mock_lit = MagicMock()

        # Mock comparison operations on these mock objects
        mock_col.__gt__.return_value = (
            MagicMock()
        )  # For `f.col(timestamp_column) > f.lit(...)`
        mock_col.__lt__.return_value = (
            MagicMock()
        )  # For `f.col(timestamp_column) < f.lit(...)`

        mock_f.col.return_value = mock_col
        mock_f.lit.return_value = mock_lit

        # Mock the filter and withColumn methods to return the DataFrame itself
        mock_df.filter.return_value = mock_df
        mock_df.withColumn.return_value = mock_df

        # Call the upload_omop_tables function (not main)
        upload_omop_tables(
            mock_spark,
            Path("measurement"),
            {
                "base_url": "jdbc:postgresql://localhost:5432/testdb",
                "user": "test_user",
                "password": "test_password",
            },
        )

        # Verify that the DataFrame was read correctly and filtered
        mock_spark.read.format.assert_called_with("parquet")
        mock_spark.read.format().load.assert_called_with("measurement/")
        mock_df.filter.assert_any_call(
            mock_f.col("measurement_date") > mock_f.lit("1900-01-01").cast("date")
        )
        mock_df.filter.assert_any_call(
            mock_f.col("measurement_date") < mock_f.lit("9999-01-01").cast("date")
        )

        # Verify that the columns were cast correctly
        mock_df.withColumn.assert_any_call(
            "concept_id", mock_f.col("concept_id").cast("integer")
        )
        mock_df.withColumn.assert_any_call(
            "measurement_date", mock_f.col("measurement_date").cast("date")
        )

        # Verify that the DataFrame was written to the database
        mock_df.repartition.assert_called_with(10)
        mock_df.repartition().write.format.assert_called_with("jdbc")
        mock_df.repartition().write.format().options.assert_called_with(
            url="jdbc:postgresql://localhost:5432/testdb",
            dbtable="measurement",
            user="test_user",
            password="test_password",
            batchsize=200000,
            queryTimeout=500,
        )
        mock_df.repartition().write.format().options().mode.assert_called_with(
            "overwrite"
        )
        mock_df.repartition().write.format().options().mode().save.assert_called_once()


if __name__ == "__main__":
    unittest.main()



from typing import Any, Dict, List

import numpy as np
import torch
from torch.nn.utils.rnn import pad_sequence
from transformers.utils import logging

from cehrgpt.data.cehrgpt_data_processor import CehrGptDataProcessor
from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer

LOG = logging.get_logger("transformers")


class CehrGptDataCollator:
    def __init__(
        self,
        tokenizer: CehrGptTokenizer,
        max_length: int,
        include_values: bool = False,
        shuffle_records: bool = False,
        include_ttv_prediction: bool = False,
        use_sub_time_tokenization: bool = False,
        include_motor_time_to_event: bool = False,
        motor_tte_vocab_size: int = 0,
        motor_num_time_pieces: int = 8,
        motor_sampling_probability: float = 0.5,
        pretraining: bool = True,
        include_demographics: bool = False,
        add_linear_prob_token: bool = False,
    ):
        self.tokenizer = tokenizer
        self.max_length = max_length

        self.vs_token_id = tokenizer.vs_token_id
        self.ve_token_id = tokenizer.ve_token_id

        self.include_values = include_values
        self.include_ttv_prediction = include_ttv_prediction
        self.use_sub_time_tokenization = use_sub_time_tokenization
        self.pretraining = pretraining
        self.include_demographics = include_demographics
        self.motor_code_cache: Dict[str, List[str]] = dict()

        # MOTOR TTE configuration
        if include_motor_time_to_event:
            assert motor_tte_vocab_size > 0, (
                f"motor_tte_vocab_size must be greater than 0 "
                f"when include_motor_time_to_event is set to True. "
                f"But motor_tte_vocab_size: {motor_tte_vocab_size} is provided"
            )

        self.include_motor_time_to_event = include_motor_time_to_event
        self.motor_tte_vocab_size = motor_tte_vocab_size
        self.motor_num_time_pieces = motor_num_time_pieces
        self.motor_time_bins = (
            self.tokenizer.get_motor_time_bins(motor_num_time_pieces)
            if self.include_motor_time_to_event
            else []
        )
        # Convert the time bins to seconds
        self.motor_time_bins = [time_bin * 86400 for time_bin in self.motor_time_bins]
        LOG.info("self.motor_time_bins: %s", self.motor_time_bins)
        if self.use_sub_time_tokenization:
            token_to_time_token_mapping = tokenizer.token_to_time_token_mapping
            if not token_to_time_token_mapping:
                raise ValueError(
                    "The token_to_time_token_mapping in CehrGptTokenizer cannot be None "
                    "when use_sub_time_tokenization is enabled"
                )
            # Create the tensors for converting time tokens to the sub time tokens
            self.time_tokens = torch.tensor(
                list(tokenizer.token_to_time_token_mapping.keys()), dtype=torch.int64
            )
            self.mapped_sub_time_tokens = torch.tensor(
                list(token_to_time_token_mapping.values()), dtype=torch.int64
            )

        self.cehrgpt_data_processor = CehrGptDataProcessor(
            tokenizer=tokenizer,
            max_length=self.max_length,
            shuffle_records=shuffle_records,
            include_ttv_prediction=include_ttv_prediction,
            include_values=include_values,
            include_motor_time_to_event=include_motor_time_to_event,
            motor_sampling_probability=motor_sampling_probability,
            pretraining=pretraining,
            add_linear_prob_token=add_linear_prob_token,
        )

    def _try_reverse_tensor(self, tensor: torch.Tensor) -> torch.Tensor:
        if not self.pretraining:
            return torch.flip(tensor, dims=[-1])
        return tensor

    @staticmethod
    def _convert_to_tensor(features: Any) -> torch.Tensor:
        if isinstance(features, torch.Tensor):
            return features
        else:
            return torch.tensor(features)

    def create_time_to_event_tensors_ultra_optimized(
        self, record: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Ultra-optimized version using advanced vectorization techniques."""
        motor_row_indices = record["motor_row_indices"]
        motor_col_indices = record["motor_col_indices"]
        motor_values = record["motor_values"]
        motor_censor_times = record["motor_censor_times"]

        if len(motor_row_indices) == 0:
            # Handle empty case - use tuples for better performance
            empty_shape = (
                0,
                self.motor_num_time_pieces,
                self.tokenizer.motor_tte_vocab_size,
            )
            record["motor_tte_times"] = np.zeros(empty_shape, dtype=np.float32)
            record["motor_tte_event_indicators"] = np.zeros(empty_shape, dtype=bool)
            record["motor_tte_masks"] = np.zeros(empty_shape, dtype=bool)
            return record

        # Convert to numpy arrays once and get dimensions
        motor_row_indices = np.asarray(motor_row_indices, dtype=np.int32)
        motor_col_indices = np.asarray(motor_col_indices, dtype=np.int32)
        motor_values = np.asarray(motor_values, dtype=np.float32)
        motor_censor_times = np.asarray(motor_censor_times, dtype=np.float32)

        n_tte_predictions = len(motor_censor_times)  # More direct than unique()
        vocab_size = self.tokenizer.motor_tte_vocab_size
        n_time_pieces = self.motor_num_time_pieces

        # Create time_vectors more efficiently without broadcasting copy
        time_vectors = np.tile(
            motor_censor_times[:, np.newaxis], (1, vocab_size)
        ).astype(np.float32)
        event_indicators = np.zeros((n_tte_predictions, vocab_size), dtype=bool)

        # Vectorized assignment (already optimal)
        time_vectors[motor_row_indices, motor_col_indices] = motor_values
        event_indicators[motor_row_indices, motor_col_indices] = True

        # Early return if no predictions
        if n_tte_predictions == 0:
            empty_shape = (0, n_time_pieces, vocab_size)
            record["motor_tte_times"] = np.zeros(empty_shape, dtype=np.float32)
            record["motor_tte_event_indicators"] = np.zeros(empty_shape, dtype=bool)
            record["motor_tte_masks"] = np.zeros(empty_shape, dtype=bool)
            return record

        # Cache motor_time_bins as numpy array to avoid repeated conversion
        if not hasattr(self, "_motor_time_bins_array"):
            self._motor_time_bins_array = np.asarray(
                self.motor_time_bins, dtype=np.float32
            )

        motor_time_bins = self._motor_time_bins_array
        start_times = motor_time_bins[:-1]
        end_times = motor_time_bins[1:]
        bin_widths = end_times - start_times  # Pre-compute bin widths

        # ELIMINATED TRANSPOSE: Compute directly in target shape (n_pred, n_bins, vocab)
        # Reshape for broadcasting in target order
        time_vectors_3d = time_vectors[:, np.newaxis, :]  # (n_pred, 1, vocab)
        event_indicators_3d = event_indicators[:, np.newaxis, :]  # (n_pred, 1, vocab)

        # Broadcast time bins to match target shape
        start_times_broadcast = start_times[np.newaxis, :, np.newaxis]  # (1, n_bins, 1)
        bin_widths_broadcast = bin_widths[np.newaxis, :, np.newaxis]  # (1, n_bins, 1)

        # Compute directly in target shape (n_pred, n_bins, vocab)
        time_diff = time_vectors_3d - start_times_broadcast
        time_in_bin = np.clip(time_diff, 0, bin_widths_broadcast)

        # Optimized mask computation
        mask = time_in_bin > 0

        # More efficient log computation with better constant
        log_constant = 1e-8  # Better numerical stability than 1e-10
        time_in_bin_log = np.where(
            mask, np.log2(np.maximum(time_in_bin, log_constant)), -np.inf
        )

        # Event indicator computation in target shape
        end_times_broadcast = motor_time_bins[1:][np.newaxis, :, np.newaxis]
        time_in_range = (time_vectors_3d >= start_times_broadcast) & (
            time_vectors_3d < end_times_broadcast
        )
        event_in_bin = event_indicators_3d & time_in_range

        # Combined mask computation
        final_mask = mask | event_in_bin

        # Direct assignment - NO TRANSPOSE NEEDED!
        record["motor_tte_times"] = time_in_bin_log
        record["motor_tte_event_indicators"] = event_in_bin
        record["motor_tte_masks"] = final_mask

        # Validation (keep as is - important for correctness)
        assert (
            sum(record["motor_tte_task_indicators"]) == n_tte_predictions
        ), f'sum(record["motor_tte_task_indicators"]) == n_tte_predictions must be true'

        # Clean up input data
        del record["motor_row_indices"]
        del record["motor_col_indices"]
        del record["motor_values"]

        return record

    def __call__(self, examples):

        if not getattr(self, "sample_packing", False):
            examples = [self.cehrgpt_data_processor.transform(_) for _ in examples]

        batch = {}

        # Assume that each example in the batch is a dictionary with 'input_ids' and 'attention_mask'
        batch_input_ids = [
            self._try_reverse_tensor(self._convert_to_tensor(example["input_ids"]))
            for example in examples
        ]

        batch_attention_mask = [
            self._try_reverse_tensor(
                self._convert_to_tensor(example["attention_mask"]).to(torch.float)
                if "attention_mask" in example
                else torch.ones_like(
                    self._convert_to_tensor(example["input_ids"]), dtype=torch.float
                )
            )
            for example in examples
        ]

        # Pad sequences to the max length in the batch
        batch["input_ids"] = self._try_reverse_tensor(
            pad_sequence(
                batch_input_ids,
                batch_first=True,
                padding_value=self.tokenizer.pad_token_id,
            ).to(torch.int64)
        )

        batch["attention_mask"] = self._try_reverse_tensor(
            pad_sequence(batch_attention_mask, batch_first=True, padding_value=0.0)
        )

        if batch["input_ids"].shape[1] > self.max_length:
            LOG.warning(
                "batch[input_ids].shape[1] > self.max_length. batch[input_ids].shape[1]: %s, self.max_length: %s",
                batch["input_ids"].shape[1],
                self.max_length,
            )
        assert batch["attention_mask"].shape[1] == batch["input_ids"].shape[1], (
            f'batch["attention_mask"].shape[1]: {batch["attention_mask"].shape[1]}, '
            f'batch["input_ids"].shape[1]: {batch["input_ids"].shape[1]}'
        )
        assert batch["input_ids"].max() < self.tokenizer.vocab_size, (
            f"batch['input_ids'].max(): {batch['input_ids'].max()} must be smaller than "
            f"self.tokenizer.vocab_size: {self.tokenizer.vocab_size}. "
            f"batch['input_ids']: {batch['input_ids']} "
        )

        batch_ages = [
            self._try_reverse_tensor(self._convert_to_tensor(example["ages"]))
            for example in examples
        ]
        # Pad sequences to the max length in the batch
        batch["ages"] = self._try_reverse_tensor(
            pad_sequence(
                batch_ages,
                batch_first=True,
                padding_value=0,
            ).to(torch.int64)
        )

        batch_epoch_times = [
            self._try_reverse_tensor(self._convert_to_tensor(example["epoch_times"]))
            for example in examples
        ]
        # Pad sequences to the max length in the batch
        batch["epoch_times"] = self._try_reverse_tensor(
            pad_sequence(
                batch_epoch_times,
                batch_first=True,
                padding_value=0,
            ).to(torch.float32)
        )

        if self.pretraining:
            batch["labels"] = torch.where(
                (batch["input_ids"] != self.tokenizer.pad_token_id)
                & batch["attention_mask"].to(torch.bool),
                batch["input_ids"],
                -100,
            )

        if self.use_sub_time_tokenization:
            time_token_indicators = torch.isin(batch["input_ids"], self.time_tokens)
            masked_tokens = batch["input_ids"].clone()
            masked_tokens[~time_token_indicators] = -1
            # Get the index of the sub_time_tokens from the time_tokens tensor
            sub_time_token_indices = torch.argmax(
                (
                    masked_tokens.unsqueeze(-1)
                    == self.time_tokens.unsqueeze(0).unsqueeze(0)
                ).to(torch.int32),
                dim=-1,
            )
            sub_time_tokens = self.mapped_sub_time_tokens[sub_time_token_indices]
            batch["time_token_indicators"] = time_token_indicators
            batch["sub_time_tokens"] = sub_time_tokens

        if self.include_ttv_prediction:
            batch_time_to_visits = [
                self._try_reverse_tensor(
                    self._convert_to_tensor(example["time_to_visits"])
                )
                for example in examples
            ]
            batch["time_to_visits"] = self._try_reverse_tensor(
                pad_sequence(
                    batch_time_to_visits, batch_first=True, padding_value=-100.0
                )
            )

        if self.include_motor_time_to_event:
            examples_with_motor_tte = [
                self.create_time_to_event_tensors_ultra_optimized(_) for _ in examples
            ]
            # print(f"Creating MOTOR TTE tensors took {time.time() - start} seconds")
            motor_tte_times = [
                self._try_reverse_tensor(
                    self._convert_to_tensor(example["motor_tte_times"])
                )
                for example in examples_with_motor_tte
            ]
            motor_tte_event_indicators = [
                self._try_reverse_tensor(
                    self._convert_to_tensor(example["motor_tte_event_indicators"])
                )
                for example in examples_with_motor_tte
            ]
            motor_tte_task_indicators = [
                self._try_reverse_tensor(
                    self._convert_to_tensor(example["motor_tte_task_indicators"])
                )
                for example in examples_with_motor_tte
            ]
            motor_tte_masks = [
                self._try_reverse_tensor(
                    self._convert_to_tensor(example["motor_tte_masks"])
                )
                for example in examples_with_motor_tte
            ]

            motor_tte_times = torch.concat(motor_tte_times, dim=0).to(torch.float32)

            # If every example in the batch only contains one visit, there would be no labels generated for MOTOR TTE
            # we only create the labels when any example has more than one visit
            if motor_tte_times.dim() <= 1:
                LOG.warning(
                    "There are no MOTOR TTE labels generated for this batch "
                    "because every example in this batch only contains one visit."
                )
            else:
                batch_size = len(examples)
                length, num_time_pieces, motor_tte_vocab_size = motor_tte_times.shape
                padded_length = batch_size - length % batch_size
                batch["motor_tte_times"] = (
                    torch.concat(
                        [
                            motor_tte_times,
                            torch.full(
                                (padded_length, num_time_pieces, motor_tte_vocab_size),
                                0.0,
                            ),
                        ],
                        dim=0,
                    )
                    .reshape((batch_size, -1, num_time_pieces, motor_tte_vocab_size))
                    .to(torch.float32)
                )

                # Motor event indicators that indicate there is an event occurred in this time interval
                batch["motor_tte_event_indicators"] = (
                    torch.concat(
                        [
                            torch.concat(motor_tte_event_indicators, dim=0).to(
                                torch.bool
                            ),
                            torch.full(
                                (padded_length, num_time_pieces, motor_tte_vocab_size),
                                False,
                            ),
                        ],
                        dim=0,
                    )
                    .reshape((batch_size, -1, num_time_pieces, motor_tte_vocab_size))
                    .to(torch.bool)
                )

                # Input to indicate whether the visit should be included for TTE predictions
                batch["motor_tte_task_indicators"] = pad_sequence(
                    motor_tte_task_indicators,
                    batch_first=True,
                    padding_value=False,
                ).to(torch.bool)

                # Motor time indicators that indicate whether there are neither clinical events nor censor events
                batch["motor_tte_masks"] = (
                    torch.concat(
                        [
                            torch.concat(motor_tte_masks, dim=0).to(torch.bool),
                            torch.full(
                                (padded_length, num_time_pieces, motor_tte_vocab_size),
                                False,
                            ),
                        ],
                        dim=0,
                    )
                    .reshape((batch_size, -1, num_time_pieces, motor_tte_vocab_size))
                    .to(torch.bool)
                )

                batch["motor_end_index"] = torch.concat(
                    [
                        torch.full((length, 1), 1, dtype=torch.int32),
                        torch.full((padded_length, 1), 0, dtype=torch.int32),
                    ]
                ).reshape((batch_size, -1))

        if self.include_values:
            batch_value_indicators = [
                self._try_reverse_tensor(
                    self._convert_to_tensor(example["value_indicators"]).to(torch.bool)
                )
                for example in examples
            ]
            batch_values = [
                self._try_reverse_tensor(self._convert_to_tensor(example["values"]))
                for example in examples
            ]
            batch["value_indicators"] = self._try_reverse_tensor(
                pad_sequence(
                    batch_value_indicators, batch_first=True, padding_value=False
                )
            )
            batch["values"] = self._try_reverse_tensor(
                pad_sequence(
                    batch_values,
                    batch_first=True,
                    padding_value=self.tokenizer.pad_value_token_id,
                ).to(torch.int64)
            )

            if batch["value_indicators"].shape[1] > self.max_length:
                LOG.warning(
                    "The total number of values in the packed sequence should be less than %s. "
                    "But the total number of tokens is: %s",
                    self.max_length,
                    batch["value_indicators"].shape[1],
                )

            if self.pretraining:
                batch["true_value_indicators"] = batch["value_indicators"].clone()
                batch["true_values"] = torch.where(
                    batch["value_indicators"], batch["values"].clone(), -100
                )

        bz = len(examples)
        if "person_id" in examples[0]:
            batch["person_id"] = (
                torch.cat(
                    [
                        self._convert_to_tensor(example["person_id"]).reshape(-1, 1)
                        for example in examples
                    ],
                    dim=0,
                )
                .to(torch.int32)
                .reshape(bz, -1)
            )

        if "index_date" in examples[0]:
            batch["index_date"] = torch.cat(
                [
                    torch.tensor(example["index_date"], dtype=torch.float64).reshape(
                        -1, 1
                    )
                    for example in examples
                ],
                dim=0,
            ).reshape(bz, -1)

        if "age_at_index" in examples[0]:
            batch["age_at_index"] = (
                torch.cat(
                    [
                        self._convert_to_tensor(example["age_at_index"]).reshape(-1, 1)
                        for example in examples
                    ],
                    dim=0,
                )
                .to(torch.float32)
                .reshape(bz, -1)
            )

        if "classifier_label" in examples[0]:
            batch["classifier_label"] = (
                torch.cat(
                    [
                        self._convert_to_tensor(example["classifier_label"]).reshape(
                            -1, 1
                        )
                        for example in examples
                    ],
                    dim=0,
                )
                .to(torch.float32)
                .reshape(bz, -1)
            )

        return batch


class SamplePackingCehrGptDataCollator(CehrGptDataCollator):
    def __init__(self, max_tokens, max_position_embeddings, *args, **kwargs):
        self.max_tokens_per_batch = max_tokens
        self.max_position_embeddings = max_position_embeddings
        self.sample_packing = True
        super(SamplePackingCehrGptDataCollator, self).__init__(*args, **kwargs)
        self.cehrgpt_data_processor.max_length = self.max_position_embeddings

    def __call__(self, examples):
        current_input_ids = []
        current_attention_mask = []
        current_ages = []
        current_epoch_times = []
        current_value_indicators = []
        current_values = []
        current_time_to_visits = []

        # MOTOR inputs
        current_motor_censor_times = []
        current_motor_row_indices = []
        current_motor_col_indices = []
        current_motor_values = []
        current_motor_tte_task_indicators = []

        # Demographics
        current_person_ids = []
        current_index_dates = []

        # Binary classification inputs
        current_prediction_ages = []
        current_labels = []

        for idx, example in enumerate(examples):
            example = self.cehrgpt_data_processor.transform(example)
            input_ids = example["input_ids"]
            # We add [END] [PAD], we want to attend to [END], adding [END] is important for sequence generation.
            # If the sequence length of the sequence is less than the context window, we add both [END][PAD], otherwise
            # we only add [PAD] token to the end of the sequence because it's not finished
            current_input_ids.extend(list(input_ids) + [self.tokenizer.pad_token_id])
            current_attention_mask.extend(np.ones_like(input_ids).tolist() + [0])

            ages = (
                example["ages"].tolist()
                if isinstance(example["ages"], torch.Tensor)
                else list(example["ages"])
            )
            current_ages.extend(ages + [max(ages)] if len(ages) > 0 else [])

            epoch_times = (
                example["epoch_times"].tolist()
                if isinstance(example["epoch_times"], torch.Tensor)
                else list(example["epoch_times"])
            )
            current_epoch_times.extend(
                epoch_times + [max(epoch_times)] if len(epoch_times) > 0 else []
            )

            if self.include_ttv_prediction:
                current_time_to_visits.extend(
                    (
                        example["time_to_visits"].tolist()
                        if isinstance(example["time_to_visits"], torch.Tensor)
                        else list(example["time_to_visits"])
                    )
                    + [-100]
                )

            if self.include_values:
                current_value_indicators.extend(
                    (
                        example["value_indicators"].tolist()
                        if isinstance(example["value_indicators"], torch.Tensor)
                        else list(example["value_indicators"])
                    )
                    + [False]
                )
                current_values.extend(
                    (
                        example["values"].tolist()
                        if isinstance(example["values"], torch.Tensor)
                        else list(example["values"])
                    )
                    + [self.tokenizer.pad_value_token_id]
                )

            if self.include_motor_time_to_event:
                current_max_motor_row_index = len(np.unique(current_motor_row_indices))
                motor_row_indices = (
                    example["motor_row_indices"].tolist()
                    if isinstance(example["motor_row_indices"], torch.Tensor)
                    else list(example["motor_row_indices"])
                )
                current_motor_row_indices.extend(
                    list(
                        map(
                            lambda offset: offset + current_max_motor_row_index,
                            motor_row_indices,
                        )
                    )
                )
                current_motor_col_indices.extend(
                    example["motor_col_indices"].tolist()
                    if isinstance(example["motor_col_indices"], torch.Tensor)
                    else list(example["motor_col_indices"])
                )
                current_motor_values.extend(
                    example["motor_values"].tolist()
                    if isinstance(example["motor_values"], torch.Tensor)
                    else list(example["motor_values"])
                )
                current_motor_censor_times.extend(
                    example["motor_censor_times"].tolist()
                    if isinstance(example["motor_censor_times"], torch.Tensor)
                    else list(example["motor_censor_times"])
                )
                current_motor_tte_task_indicators.extend(
                    (
                        example["motor_tte_task_indicators"].tolist()
                        if isinstance(
                            example["motor_tte_task_indicators"], torch.Tensor
                        )
                        else list(example["motor_tte_task_indicators"])
                    )
                    + [False]
                )

            if "person_id" in example:
                current_person_ids.append(example["person_id"])

            if "index_date" in example:
                current_index_dates.append(example["index_date"])

            if "age_at_index" in example:
                current_prediction_ages.append(example["age_at_index"])

            if "classifier_label" in example:
                current_labels.append(example["classifier_label"])

        if len(current_input_ids) > self.max_tokens_per_batch:
            LOG.warning(
                "The total number of tokens in the packed sequence should be less than %s. "
                "But the total number of tokens is: %s",
                self.max_tokens_per_batch,
                len(current_input_ids),
            )

        packed_example = {
            "input_ids": current_input_ids,
            "attention_mask": current_attention_mask,
            "ages": current_ages,
            "epoch_times": current_epoch_times,
        }

        if self.include_ttv_prediction:
            packed_example.update({"time_to_visits": current_time_to_visits})

        if self.include_values:
            packed_example.update(
                {"value_indicators": current_value_indicators, "values": current_values}
            )
        if self.include_motor_time_to_event:
            packed_example.update(
                {
                    "motor_censor_times": current_motor_censor_times,
                    "motor_row_indices": current_motor_row_indices,
                    "motor_col_indices": current_motor_col_indices,
                    "motor_values": current_motor_values,
                    "motor_tte_task_indicators": current_motor_tte_task_indicators,
                }
            )

        if current_labels:
            packed_example.update(
                {
                    "person_id": current_person_ids,
                    "index_date": current_index_dates,
                    "age_at_index": current_prediction_ages,
                    "classifier_label": current_labels,
                }
            )
        # print(f"Packing examples took {time.time() - start} seconds")
        return super().__call__([packed_example])



from __future__ import annotations

import collections
import os
from typing import Any, Dict, Iterable, Optional, Set, Union

import polars as pl
from datasets import Dataset


# Adapted from femr.ontology
def _get_all_codes_map(batch: Dict[str, Any]) -> Dict[str, Any]:
    result = set()
    for concept_ids in batch["concept_ids"]:
        for concept_id in concept_ids:
            if concept_id.isnumeric():
                result.add(concept_id)
    return {"unique_concept_ids": list(result)}


class Ontology:
    def __init__(self, vocab_path: str):
        """Create an Ontology from an Athena download and an optional meds Code Metadata structure.

        NOTE: This is an expensive operation.
        It is recommended to create an ontology once and then save/load it as necessary.
        """
        # Load from code metadata
        self.parents_map: Dict[str, Set[str]] = collections.defaultdict(set)
        self.concept_vocabulary_map: Dict[str, str] = collections.defaultdict(str)
        self.concept_domain_map: Dict[str, str] = collections.defaultdict(str)

        # Load from the athena path ...
        concept = pl.scan_parquet(os.path.join(vocab_path, "concept/*parquet"))
        vocabulary_id_col = pl.col("vocabulary_id")
        concept_id_col = pl.col("concept_id").cast(pl.String)
        domain_id_col = pl.col("domain_id").cast(pl.String)

        processed_concepts = (
            concept.select(
                concept_id_col,
                domain_id_col,
                vocabulary_id_col,
                pl.col("standard_concept").is_null(),
            )
            .collect()
            .rows()
        )

        non_standard_concepts = set()

        for concept_id, domain_id, vocabulary_id, is_non_standard in processed_concepts:
            # We don't want to override code metadata
            if concept_id not in self.concept_vocabulary_map:
                self.concept_vocabulary_map[concept_id] = vocabulary_id

            if concept_id not in self.concept_domain_map:
                self.concept_domain_map[concept_id] = domain_id
            # We don't want to override code metadata
            if is_non_standard:
                non_standard_concepts.add(concept_id)

        relationship = pl.scan_parquet(
            os.path.join(vocab_path, "concept_relationship/*parquet")
        )
        relationship_id = pl.col("relationship_id")
        relationship = relationship.filter(
            relationship_id == "Maps to",
            pl.col("concept_id_1") != pl.col("concept_id_2"),
        )
        for concept_id_1, concept_id_2 in (
            relationship.select(
                pl.col("concept_id_1").cast(pl.String),
                pl.col("concept_id_2").cast(pl.String),
            )
            .collect()
            .rows()
        ):
            if concept_id_1 in non_standard_concepts:
                self.parents_map[concept_id_1].add(concept_id_2)

        ancestor = pl.scan_parquet(
            os.path.join(vocab_path, "concept_ancestor/*parquet")
        )
        ancestor = ancestor.filter(pl.col("min_levels_of_separation") == 1)
        for concept_id, parent_concept_id in (
            ancestor.select(
                pl.col("descendant_concept_id").cast(pl.String),
                pl.col("ancestor_concept_id").cast(pl.String),
            )
            .collect()
            .rows()
        ):
            self.parents_map[concept_id].add(parent_concept_id)
        self.all_parents_map: Dict[str, Set[str]] = {}

    def get_domain(self, concept_id: Union[str, int]) -> Optional[str]:
        return self.concept_domain_map.get(str(concept_id), None)

    def prune_to_dataset(
        self,
        dataset: Dataset,
        remove_ontologies: Set[str] = set(),
        num_proc: int = 4,
        batch_size: int = 1024,
    ) -> None:
        mapped_dataset = dataset.map(
            _get_all_codes_map,
            batched=True,
            batch_size=batch_size,
            remove_columns=dataset.column_names,
            num_proc=num_proc,
        )
        valid_concept_ids = set(mapped_dataset["unique_concept_ids"])
        all_parents = set()
        for concept_id in valid_concept_ids:
            all_parents |= self.get_all_parents(concept_id)

        def is_valid(c: str):
            ontology = self.concept_vocabulary_map.get(c, "")
            return (c in valid_concept_ids) or (
                (ontology not in remove_ontologies) and (c in all_parents)
            )

        concept_ids = set(self.parents_map.keys())
        for concept_id in concept_ids:
            m: Any
            if is_valid(concept_id):
                for m in (self.parents_map, self.concept_vocabulary_map):
                    m[concept_id] = {a for a in m[concept_id] if is_valid(a)}
            else:
                for m in (self.parents_map, self.concept_vocabulary_map):
                    if concept_id in m:
                        del m[concept_id]

        self.all_parents_map = {}

        # Prime the pump
        for concept_id in self.parents_map.keys():
            self.get_all_parents(concept_id)

    def get_parents(self, code: str) -> Iterable[str]:
        """Get the parents for a given code."""
        return self.parents_map.get(code, set())

    def get_all_parents(self, code: str) -> Set[str]:
        """Get all parents, including through the ontology."""
        if code not in self.all_parents_map:
            result = {code}
            for parent in self.parents_map.get(code, set()):
                result |= self.get_all_parents(parent)
            self.all_parents_map[code] = result

        return self.all_parents_map[code]



import dataclasses
from typing import List, Literal, Optional

from cehrgpt.models.gpt2 import ACT2FN


@dataclasses.dataclass
class CehrGPTArguments:
    """Arguments pertaining to what data we are going to input our model for training and eval."""

    patient_splits_path: Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "Path to the patient splits file"},
    )
    tokenized_dataset_name: Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "The name of the dataset to use for tokenization."},
    )
    tokenized_full_dataset_path: Optional[str] = dataclasses.field(
        default=None,
        metadata={
            "help": "The path to the tokenized dataset created for the full population"
        },
    )
    activation_function: Literal[tuple(ACT2FN.keys())] = dataclasses.field(
        default="gelu_new",
        metadata={"help": "The activation function to use"},
    )
    decoder_mlp: Literal["GPT2MLP", "LlamaMLP"] = dataclasses.field(
        default="GPT2MLP",
        metadata={"help": "The decoder MLP architecture"},
    )
    include_inpatient_hour_token: Optional[bool] = dataclasses.field(
        default=True,
        metadata={"help": "Include inpatient hour token"},
    )
    include_demographics: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to always include the demographics for the long sequences that are longer than the model context window."
        },
    )
    continue_pretrain: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to continue to pretrain cehrgpt on the new dataset"
        },
    )
    pretrained_embedding_path: Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "The path to the concept pretrained embeddings"},
    )
    retrain_with_full: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to retrain the model on the full set after early stopping"
        },
    )
    expand_tokenizer: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to expand the tokenizer for fine-tuning."
        },
    )
    few_shot_predict: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to use a few shots to train the model"
        },
    )
    n_shots: Optional[int] = dataclasses.field(
        default=128,
        metadata={"help": "The number of examples from the training set."},
    )
    hyperparameter_tuning: Optional[bool] = dataclasses.field(
        default=False,
        metadata={"help": "A flag to indicate if we want to do hyperparameter tuning."},
    )
    hyperparameter_tuning_is_grid: Optional[bool] = dataclasses.field(
        default=True,
        metadata={"help": "A flag to indicate if we want to do hyperparameter tuning."},
    )
    hyperparameter_tuning_percentage: Optional[float] = dataclasses.field(
        default=0.1,
        metadata={
            "help": "The percentage of the train/val will be use for hyperparameter tuning."
        },
    )
    n_trials: Optional[int] = dataclasses.field(
        default=10,
        metadata={
            "help": "The number of trails will be use for hyperparameter tuning."
        },
    )
    hyperparameter_batch_sizes: Optional[List[int]] = dataclasses.field(
        default_factory=lambda: [4, 8, 16],
        metadata={"help": "Hyperparameter search batch sizes"},
    )
    hyperparameter_num_train_epochs: Optional[List[int]] = dataclasses.field(
        default_factory=lambda: [10],
        metadata={"help": "Hyperparameter search num_train_epochs"},
    )
    hyperparameter_learning_rates: Optional[List[int]] = dataclasses.field(
        default_factory=lambda: [1e-5],
        metadata={"help": "Hyperparameter search learning rates"},
    )
    hyperparameter_weight_decays: Optional[List[int]] = dataclasses.field(
        default_factory=lambda: [1e-2],
        metadata={"help": "Hyperparameter search learning rates"},
    )
    causal_sfm: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether the GPT conforms to the causal Standard Fairness Model"
        },
    )
    demographics_size: Optional[int] = dataclasses.field(
        default=4,
        metadata={
            "help": "The number of demographics tokens in the patient sequence "
            "It defaults to 4, assuming the demographics tokens follow this pattern [Year][Age][Gender][Race]"
        },
    )
    drop_long_sequences: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "The lower bound of the learning rate range for hyperparameter tuning."
        },
    )
    next_token_prediction_loss_weight: float = dataclasses.field(
        default=1.0, metadata={"help": "The weight of the next token prediction loss"}
    )
    lab_token_penalty: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to use lab token loss penalty."
        },
    )
    lab_token_loss_weight: Optional[float] = dataclasses.field(
        default=1.0,
        metadata={"help": "lab_token_loss_weight penalty co-efficient"},
    )
    value_prediction_loss_weight: Optional[float] = dataclasses.field(
        default=1.0,
        metadata={"help": "The weight of the value prediction loss"},
    )
    entropy_penalty: Optional[bool] = dataclasses.field(
        default=False,
        metadata={"help": "A flag to indicate whether we want to use entropy penalty."},
    )
    entropy_penalty_alpha: Optional[float] = dataclasses.field(
        default=0.01,
        metadata={"help": "Entropy penalty co-efficient"},
    )
    n_pretrained_embeddings_layers: Optional[int] = dataclasses.field(
        default=2,
        metadata={
            "help": "The number of feed forward layers for transforming pretrained embeddings to internal embeddings"
        },
    )
    meds_repartition: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to repartition the meds train tune sets"
        },
    )
    use_early_stopping: Optional[bool] = dataclasses.field(
        default=True,
        metadata={"help": "A flag to indicate whether we want to use early stopping."},
    )
    early_stopping_threshold: Optional[float] = dataclasses.field(
        default=0.01,
        metadata={
            "help": "A threshold to denote how much the specified metric must improve to satisfy early stopping conditions."
        },
    )
    inner_dim: Optional[int] = dataclasses.field(
        default=None,
        metadata={"help": "The dimensionality of the hidden layer"},
    )
    apply_rotary: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to use rotary encoder layers"
        },
    )
    sample_packing: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to use sample packing for efficient training."
        },
    )
    max_tokens_per_batch: int = dataclasses.field(
        default=16384, metadata={"help": "Maximum number of tokens in each batch"}
    )
    include_motor_time_to_event: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to include the motor time to events"
        },
    )
    num_motor_tasks: Optional[int] = dataclasses.field(
        default=10000,
        metadata={"help": "The number of max MOTOR tasks"},
    )
    motor_time_to_event_weight: Optional[float] = dataclasses.field(
        default=1.0,
        metadata={"help": "The MOTOR time to event loss weight"},
    )
    motor_num_time_pieces: Optional[int] = dataclasses.field(
        default=8,
        metadata={
            "help": "The number of times each motor_num_time_pieces piece has to be"
        },
    )
    motor_use_ontology: Optional[bool] = dataclasses.field(
        default=False,
        metadata={
            "help": "A flag to indicate whether we want to use motor_use_ontology"
        },
    )
    motor_sampling_probability: Optional[float] = dataclasses.field(
        default=0.0,
        metadata={"help": "A flag to indicate whether we want to use sample packing"},
    )
    vocab_dir: Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "The directory where the concept data is stored."},
    )
    average_over_sequence: bool = dataclasses.field(
        default=False,
        metadata={"help": "Whether or not to average tokens per sequence"},
    )
    apply_entropy_filter: Optional[bool] = dataclasses.field(
        default=False,
        metadata={"help": "A flag to indicate whether we want to use entropy filter."},
    )
    min_prevalence: Optional[float] = dataclasses.field(
        default=1 / 1000,
        metadata={"help": "The min_prevalence to keep the concepts in the tokenizer"},
    )
    class_weights: Optional[List[int]] = dataclasses.field(
        default=None,
        metadata={"help": "The class weights for training"},
    )
    negative_sampling_probability: Optional[float] = dataclasses.field(
        default=None,
        metadata={
            "help": "The probability of negative samples will be included in the training data"
        },
    )
    num_of_trajectories_per_sample: Optional[int] = dataclasses.field(
        default=1,
        metadata={"help": "The number of trajectories per sample"},
    )
    generation_input_length: Optional[int] = dataclasses.field(
        default=1024,
        metadata={"help": "The length of the input sequence"},
    )
    generation_max_new_tokens: Optional[int] = dataclasses.field(
        default=1024,
        metadata={"help": "The maximum number of tokens in the generation sequence"},
    )



import datetime
import os
import random
import uuid
from typing import Any, Dict, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
import torch
from cehrbert.runners.runner_util import load_parquet_as_dataset
from transformers import GenerationConfig
from transformers.utils import is_flash_attn_2_available, logging

from cehrgpt.cehrgpt_args import create_inference_base_arg_parser
from cehrgpt.generation.omop_converter_batch import START_TOKEN_SIZE
from cehrgpt.gpt_utils import construct_age_sequence, get_cehrgpt_output_folder
from cehrgpt.models.hf_cehrgpt import CEHRGPT2LMHeadModel
from cehrgpt.models.special_tokens import END_TOKEN
from cehrgpt.models.tokenization_hf_cehrgpt import (
    NA,
    CehrGptTokenizer,
    is_valid_valid_bin,
)

LOG = logging.get_logger("transformers")


def normalize_value(
    seq: Sequence[str],
    values: Sequence[str],
    tokenizer: CehrGptTokenizer,
) -> Tuple[
    Sequence[str],
    Optional[Sequence[Optional[int]]],
    Optional[Sequence[Optional[float]]],
    Optional[Sequence[Optional[str]]],
    Optional[Sequence[str]],
]:
    concepts = []
    number_as_values = []
    concept_as_values = []
    is_numeric_types = []
    units = []
    for concept, value in zip(seq, values):
        if concept == END_TOKEN:
            break
        number_as_value = None
        concept_as_value = value if value and value.isnumeric() else None
        is_numeric_type = 0
        unit = NA
        # If concept is numeric, we expect the next token to be a value bin
        if is_valid_valid_bin(value):
            converted_value, unit = tokenizer.denormalize(concept, value)
            if isinstance(converted_value, float):
                number_as_value = converted_value
                is_numeric_type = 1

        concepts.append(concept)
        number_as_values.append(number_as_value)
        concept_as_values.append(concept_as_value)
        is_numeric_types.append(is_numeric_type)
        units.append(unit)

    return (
        concepts,
        is_numeric_types,
        number_as_values,
        concept_as_values,
        units,
    )


def generate_single_batch(
    model: CEHRGPT2LMHeadModel,
    cehrgpt_tokenizer: CehrGptTokenizer,
    prompts: torch.Tensor,
    max_length: int,
    ages: Optional[torch.Tensor] = None,
    values: Optional[torch.Tensor] = None,
    value_indicators: Optional[torch.Tensor] = None,
    max_new_tokens: Optional[int] = None,
    mini_num_of_concepts=1,
    top_p=0.95,
    top_k=50,
    temperature=1.0,
    repetition_penalty=1.0,
    num_beams=1,
    num_beam_groups=1,
    epsilon_cutoff=0.0,
    device: Any = "cpu",
) -> Dict[str, Any]:
    with torch.no_grad():
        generation_config = GenerationConfig(
            repetition_penalty=repetition_penalty,
            max_new_tokens=max_new_tokens,
            max_length=max_length,
            min_length=mini_num_of_concepts,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            bos_token_id=model.generation_config.bos_token_id,
            eos_token_id=model.generation_config.eos_token_id,
            pad_token_id=model.generation_config.pad_token_id,
            do_sample=True,
            use_cache=True,
            return_dict_in_generate=True,
            output_attentions=False,
            output_hidden_states=False,
            output_scores=False,
            renormalize_logits=True,
            num_beams=num_beams,
            num_beam_groups=num_beam_groups,
            epsilon_cutoff=epsilon_cutoff,
        )

        batched_prompts = prompts.to(device)
        if ages is not None:
            ages = ages.to(device)
        if values is not None:
            values = values.to(device)
        if value_indicators is not None:
            value_indicators = value_indicators.to(device)

        results = model.generate(
            inputs=batched_prompts,
            ages=ages,
            values=values,
            value_indicators=value_indicators,
            generation_config=generation_config,
            cehrgpt_tokenizer=cehrgpt_tokenizer,
        )

    sequences = [
        cehrgpt_tokenizer.decode(seq.cpu().numpy(), skip_special_tokens=False)
        for seq in results.sequences
    ]
    if results.sequence_vals is not None:
        values = [
            cehrgpt_tokenizer.decode_value(
                values.cpu().numpy(), skip_special_tokens=False
            )
            for values in results.sequence_vals
        ]
    else:
        values = np.zeros_like(sequences)
        values.fill(NA)
    if results.sequence_val_masks is not None:
        value_indicators = results.sequence_val_masks.cpu().numpy()
    else:
        value_indicators = np.zeros_like(sequences, dtype=np.int32).astype(bool)
    return {
        "sequences": sequences,
        "values": values,
        "value_indicators": value_indicators,
    }


def main(args):
    if torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    cehrgpt_tokenizer = CehrGptTokenizer.from_pretrained(args.tokenizer_folder)
    cehrgpt_model = (
        CEHRGPT2LMHeadModel.from_pretrained(
            args.model_folder,
            attn_implementation=(
                "flash_attention_2" if is_flash_attn_2_available() else "eager"
            ),
        )
        .eval()
        .to(device)
    )

    cehrgpt_model.generation_config.pad_token_id = cehrgpt_tokenizer.pad_token_id
    cehrgpt_model.generation_config.eos_token_id = cehrgpt_tokenizer.end_token_id
    cehrgpt_model.generation_config.bos_token_id = cehrgpt_tokenizer.end_token_id

    folder_name = get_cehrgpt_output_folder(args, cehrgpt_tokenizer)
    output_folder_name = os.path.join(
        args.output_folder, folder_name, "generated_sequences"
    )

    if not os.path.exists(output_folder_name):
        os.makedirs(output_folder_name)

    # Determine whether we will use the demographics with the long sequences
    max_seq_allowed = (
        cehrgpt_model.config.n_positions
        if args.drop_long_sequences
        else np.iinfo(np.int32).max
    )

    LOG.info(f"Loading tokenizer at {args.model_folder}")
    LOG.info(f"Loading model at {args.model_folder}")
    LOG.info(f"Write sequences to {output_folder_name}")
    LOG.info(f"Context window {args.context_window}")
    LOG.info(f"Max sequence allowed {max_seq_allowed}")
    LOG.info(f"Temperature {args.temperature}")
    LOG.info(f"Repetition Penalty {args.repetition_penalty}")
    LOG.info(f"Sampling Strategy {args.sampling_strategy}")
    LOG.info(f"Num beam {args.num_beams}")
    LOG.info(f"Num beam groups {args.num_beam_groups}")
    LOG.info(f"Epsilon cutoff {args.epsilon_cutoff}")
    LOG.info(f"Top P {args.top_p}")
    LOG.info(f"Top K {args.top_k}")
    LOG.info(f"Loading demographic_info at {args.demographic_data_path}")
    LOG.info(f"MEDS format: {args.meds_format}")

    dataset = load_parquet_as_dataset(args.demographic_data_path)
    total_rows = len(dataset)

    num_of_batches = args.num_of_patients // args.batch_size + 1
    sequence_to_flush = []
    current_person_id = 1
    prompt_size = 2 if args.meds_format else START_TOKEN_SIZE
    for i in range(num_of_batches):
        LOG.info(f"{datetime.datetime.now()}: Batch {i} started")

        # Randomly pick demographics from the existing population
        random_prompts = []
        random_prompt_ages = []
        iter = 0
        while len(random_prompts) < args.batch_size:
            for row in dataset.select(
                random.sample(range(total_rows), k=args.batch_size)
            ):
                if (
                    args.min_num_of_concepts
                    <= len(row["concept_ids"])
                    <= max_seq_allowed
                ):
                    prompt = row["concept_ids"][:prompt_size]
                    random_prompts.append(cehrgpt_tokenizer.encode(prompt))
                    random_prompt_ages.append(construct_age_sequence(prompt))
                iter += 1
                if not random_prompts and iter > 10:
                    raise RuntimeError(
                        f"The length of concept_ids in {args.demographic_data_path} does not qualify!"
                    )

        # Make sure the batch does not exceed batch_size
        batch_sequences = generate_single_batch(
            cehrgpt_model,
            cehrgpt_tokenizer,
            torch.tensor(random_prompts[: args.batch_size]),
            ages=torch.tensor(random_prompt_ages[: args.batch_size]),
            max_length=args.context_window,
            mini_num_of_concepts=args.min_num_of_concepts,
            top_p=args.top_p,
            top_k=args.top_k,
            temperature=args.temperature,
            repetition_penalty=args.repetition_penalty,
            num_beams=args.num_beams,
            num_beam_groups=args.num_beam_groups,
            epsilon_cutoff=args.epsilon_cutoff,
            device=device,
        )

        # Clear the cache
        torch.cuda.empty_cache()

        for concept_ids, value_indicators, values in zip(
            batch_sequences["sequences"],
            batch_sequences["value_indicators"],
            batch_sequences["values"],
        ):
            (
                concept_ids,
                is_numeric_types,
                number_as_values,
                concept_as_values,
                units,
            ) = normalize_value(concept_ids, values, cehrgpt_tokenizer)
            output = {"concept_ids": concept_ids, "person_id": current_person_id}
            if is_numeric_types is not None:
                output["is_numeric_types"] = is_numeric_types
            if number_as_values is not None:
                output["number_as_values"] = number_as_values
            if concept_as_values is not None:
                output["concept_as_values"] = concept_as_values
            if value_indicators is not None:
                output["concept_value_masks"] = value_indicators
            if units is not None:
                output["units"] = units

            sequence_to_flush.append(output)
            current_person_id += 1

        if len(sequence_to_flush) >= args.buffer_size:
            LOG.info(f"{datetime.datetime.now()}: Flushing to the Disk at Batch {i}")
            pd.DataFrame(
                sequence_to_flush,
                columns=[
                    "concept_ids",
                    "person_id",
                    "is_numeric_types",
                    "number_as_values",
                    "concept_as_values",
                    "concept_value_masks",
                    "units",
                ],
            ).to_parquet(os.path.join(output_folder_name, f"{uuid.uuid4()}.parquet"))
            sequence_to_flush.clear()

    if len(sequence_to_flush) > 0:
        LOG.info(f"{datetime.datetime.now()}: Flushing to the Disk at Final Batch")
        pd.DataFrame(
            sequence_to_flush,
            columns=[
                "concept_ids",
                "person_id",
                "is_numeric_types",
                "number_as_values",
                "concept_as_values",
                "concept_value_masks",
                "units",
            ],
        ).to_parquet(os.path.join(output_folder_name, f"{uuid.uuid4()}-last.parquet"))


def create_arg_parser():
    base_arg_parser = create_inference_base_arg_parser(
        description="Arguments for generating patient sequences"
    )
    base_arg_parser.add_argument(
        "--num_of_patients",
        dest="num_of_patients",
        action="store",
        type=int,
        help="The number of patients that will be generated",
        required=True,
    )
    base_arg_parser.add_argument(
        "--demographic_data_path",
        dest="demographic_data_path",
        action="store",
        help="The path for your concept_path",
        required=True,
    )
    base_arg_parser.add_argument(
        "--drop_long_sequences",
        dest="drop_long_sequences",
        action="store_true",
    )
    base_arg_parser.add_argument(
        "--meds_format",
        dest="meds_format",
        action="store_true",
    )
    return base_arg_parser


if __name__ == "__main__":
    main(create_arg_parser().parse_args())



CONDITION_ERA_QUERY = """
-- Common Table Expression for cteConditionTarget
WITH cteConditionTarget AS (
  SELECT
    co.CONDITION_OCCURRENCE_ID,
    co.PERSON_ID,
    co.CONDITION_CONCEPT_ID,
    co.CONDITION_TYPE_CONCEPT_ID,
    co.CONDITION_START_DATE,
    COALESCE(co.CONDITION_END_DATE, DATE_ADD(co.CONDITION_START_DATE, 1)) AS CONDITION_END_DATE
  FROM CONDITION_OCCURRENCE co
),

-- Common Table Expression for cteEndDates
cteEndDates AS (
  SELECT
    PERSON_ID,
    CONDITION_CONCEPT_ID,
    DATE_ADD(EVENT_DATE, -30) AS END_DATE
  FROM (
    SELECT
      PERSON_ID,
      CONDITION_CONCEPT_ID,
      EVENT_DATE,
      EVENT_TYPE,
      MAX(START_ORDINAL) OVER (PARTITION BY PERSON_ID, CONDITION_CONCEPT_ID ORDER BY EVENT_DATE, EVENT_TYPE ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW) AS START_ORDINAL,
      ROW_NUMBER() OVER (PARTITION BY PERSON_ID, CONDITION_CONCEPT_ID ORDER BY EVENT_DATE, EVENT_TYPE) AS OVERALL_ORD
    FROM (
      -- Select the start dates, assigning a row number to each
      SELECT
        PERSON_ID,
        CONDITION_CONCEPT_ID,
        CONDITION_START_DATE AS EVENT_DATE,
        -1 AS EVENT_TYPE,
        ROW_NUMBER() OVER (PARTITION BY PERSON_ID, CONDITION_CONCEPT_ID ORDER BY CONDITION_START_DATE) AS START_ORDINAL
      FROM cteConditionTarget

      UNION ALL

      -- Pad the end dates by 30 to allow a grace period for overlapping ranges
      SELECT
        PERSON_ID,
        CONDITION_CONCEPT_ID,
        DATE_ADD(CONDITION_END_DATE, 30) AS EVENT_DATE,
        1 AS EVENT_TYPE,
        NULL
      FROM cteConditionTarget
    ) RAWDATA
  ) E
  WHERE (2 * E.START_ORDINAL) - E.OVERALL_ORD = 0
),

-- Common Table Expression for cteConditionEnds
cteConditionEnds AS (
  SELECT
    c.PERSON_ID,
    c.CONDITION_CONCEPT_ID,
    c.CONDITION_TYPE_CONCEPT_ID,
    c.CONDITION_START_DATE,
    MIN(e.END_DATE) AS ERA_END_DATE
  FROM cteConditionTarget c
  JOIN cteEndDates e ON c.PERSON_ID = e.PERSON_ID AND c.CONDITION_CONCEPT_ID = e.CONDITION_CONCEPT_ID AND e.END_DATE >= c.CONDITION_START_DATE
  GROUP BY
    c.PERSON_ID,
    c.CONDITION_CONCEPT_ID,
    c.CONDITION_TYPE_CONCEPT_ID,
    c.CONDITION_START_DATE
)

SELECT
  ROW_NUMBER() OVER (ORDER BY CONDITION_CONCEPT_ID) AS condition_era_id,
  PERSON_ID,
  CONDITION_CONCEPT_ID,
  MIN(CONDITION_START_DATE) AS CONDITION_ERA_START_DATE,
  ERA_END_DATE AS CONDITION_ERA_END_DATE,
  COUNT(*) AS CONDITION_OCCURRENCE_COUNT
FROM cteConditionEnds
GROUP BY
  PERSON_ID,
  CONDITION_CONCEPT_ID,
  CONDITION_TYPE_CONCEPT_ID,
  ERA_END_DATE
ORDER BY
  PERSON_ID,
  CONDITION_CONCEPT_ID
"""
